import uuid
from pathlib import Path
from functools import wraps
import time
import json
import hashlib
from typing import TYPE_CHECKING,TypeVar,Dict, List, Optional, Union,Any,Tuple,get_type_hints,Annotated,get_args,Callable
import typing
import inspect
import pydantic
import sys
import traceback
import io
from enum import Enum
from byzerllm.utils.types import VLLMStreamServer, BlockVLLMStreamServer,StreamOutputs,SingleOutput,SingleOutputMeta,BlockBinaryStreamServer

T = TypeVar("T")

def print_flush(*args, **kwargs):
    print(*args, **kwargs, flush=True)

import signal
from contextlib import contextmanager
class TimeoutException(Exception):
    pass

def timeout_handler(signum, frame):
    raise TimeoutException()

@contextmanager
def timeout(duration: float):
    signal.signal(signal.SIGALRM, timeout_handler)
    signal.alarm(duration)
    try:
        yield
    finally:
        signal.alarm(0)

def timeit(func):
    """
    Decorator to time a function.
    """

    @wraps(func)
    def inner(*args, **kwargs):
        start_time = time.monotonic()
        ret = func(*args, **kwargs)
        time_taken = time.monotonic() - start_time
        print(f"{func} took {time_taken} s to complete",flush=True)
        return ret

    return inner

def generate_instruction_from_history(ins:str,his:List[Dict[str,str]],role_mapping:Dict[str,str]={        
        "user":"User",        
        "assistant":"Assistant",
    }):

    new_his = []    
    for item in his:
        if item["role"] == "system":
            new_his.append(item["content"])
            continue        
        new_his.append(f"{role_mapping[item['role']]}:{item['content']}")            

    # here we should make sure the user build the conversation string manually also
    # works. This means if the user do not provide  the history, then
    # we should treat ins as conversation string which the user build manually
    if len(new_his) > 0 and ins != "":
        new_his.append(f"{role_mapping['user']}:{ins}")
        new_his.append(f"{role_mapping['assistant']}:")

    if len(new_his) > 0 and ins == "":
        new_his.append(f"{role_mapping['assistant']}:")            
    
    if len(new_his) == 0:
        new_his.append(ins)    

    fin_ins = "\n".join(new_his)
    return fin_ins  

def compute_max_new_tokens(tokens,max_length:int):
    input_length = tokens["input_ids"].shape[1]
    max_new_tokens = max_length - input_length
    if max_new_tokens <= 0:
        raise Exception(f"Input is too long ({input_length}). Try to reduce the length of history or use a larger `max_length` value (now:{max_length})")
    return max_new_tokens

def tokenize_string(tokenizer, key: str) -> Union[int, List[int]]:
    """Tokenize a string using a tokenizer.

    Args:
        tokenizer (PreTrainedTokenizer): Tokenizer to use.
        key (str): String to tokenize.
    """
    token_ids = tokenizer.encode(key, add_special_tokens=False)
    return token_ids

def tokenize_stopping_sequences_where_needed(
    tokenizer,
    stopping_sequences: List[Union[str, int, List[int]]],
) -> List[Union[List[int], int]]:
    """If any sequence is a string, tokenize it.

    Args:
        tokenizer (PreTrainedTokenizer): Tokenizer to use.
        stopping_sequences (List[Union[str, int, List[int]]]): Stopping sequences to
            tokenize. Can be ids, sequences of ids or strings.
    """
    if not stopping_sequences:
        return None
    return [
        tokenize_string(tokenizer, sequence) if isinstance(sequence, str) else sequence
        for sequence in stopping_sequences
    ]

def  tokenize_stopping_sequences(tokenizer,stop_words):
    stop_words_ids = []
    for stop_word in stop_words:
        w = tokenize_string(tokenizer, stop_word)
        # remove the first token which is empty token 
        # this should work for only llama model
        # if w[0] == 29871 and tokenizer.decode([w[0]],skip_special_tokens=False) == "":
        #     w = w[1:]
        stop_words_ids.append(w)    
    return stop_words_ids



def load_json_str(json_str:str):        
    return json.loads(json_str) 


def generate_file_md5(file_path: str) -> str:
    md5_hash = hashlib.md5()
    with open(file_path, 'rb') as f:
        for chunk in iter(lambda: f.read(4096), b''):
            md5_hash.update(chunk)
    return md5_hash.hexdigest()

def generate_str_md5(s: str) -> str:
    md5_hash = hashlib.md5()
    md5_hash.update(s.encode("utf-8"))
    return md5_hash.hexdigest() 

        
def get_type_name(t):
    name = str(t)
    if "list" in name or "dict" in name:
        return name
    else:
        return t.__name__
    
def is_annotated_type(hint):
    if hasattr(typing, '_AnnotatedAlias'):  # Python 3.9 and later
        return isinstance(hint, typing._AnnotatedAlias)
    elif hasattr(typing, '_SpecialForm'):  # Python versions before 3.9
        # Check if it's a _SpecialForm and its name is 'Annotated'
        return isinstance(hint, typing._SpecialForm) and hint.__name__ == 'Annotated'
    else:
        return False    
    
def serialize_function_to_json(func):
    if isinstance(func, str):
        return func
    
    signature = inspect.signature(func)
    type_hints = get_type_hints(func)

    # return_type = type_hints.get('return', 'void')
    # if return_type is None:
    #     return_type_str = 'void'
    # else:
    #     return_type_str = return_type.__name__

    function_info = {
        "name": func.__name__,
        "description": func.__doc__,
        "parameters": {
            "type": "object",
            "properties": {}
        },
        # "returns": return_type_str
    }

    for name, parameter in signature.parameters.items():
        param_type = get_type_name(type_hints.get(name, type(None)))
        param_annotated= func.__annotations__.get(name, '')

        function_info["parameters"]["properties"][name]  = {}
        properties = function_info["parameters"]["properties"][name] 

        
        if is_annotated_type(param_annotated):
            _, *metadata = get_args(param_annotated)
        else:
            metadata = []  
   
        param_desc = ""
        for meta in metadata:
            if isinstance(meta, str):
                param_desc = meta 
            if isinstance(meta, Dict):
                param_desc = meta.get("description", "")
                if "enum" in meta:
                    properties["enum"] = meta["enum"]

        properties["type"] = param_type
        properties["description"] = param_desc
        
        if parameter.default is not inspect.Parameter.empty:
            properties["default"] = parameter.default                            

    return json.dumps(function_info,ensure_ascii=False, indent=2)


class FunctionCall(pydantic.BaseModel):
    '''
    函数名称和函数参数列表
    '''        
    name: str = pydantic.Field(description="函数名")
    arguments: Dict[str,Any] = pydantic.Field(description="函数参数")

class FunctionCallWrapper(pydantic.BaseModel):    
    function: FunctionCall = pydantic.Field(description="函数调用")

class FunctionCallList(pydantic.BaseModel):
    '''
    函数调用列表    
    '''
    tool_calls: List[FunctionCallWrapper] = pydantic.Field(description="函数调用列表")
    id: str = pydantic.Field(description="工具调用的唯一标识符,无需生成")
    type: str = pydantic.Field("function",description="工具调用的类型，固定为 function，无需生成")

FUNCTION_CALLING_SCHEMA = FunctionCallList.schema_json(ensure_ascii=False, indent=2) 


def exec_capture_output(code: str,target_names:Dict[str,Any]={}) -> Tuple[int,str,Any]:
    buffer = io.StringIO()
    sys.stdout = buffer
    sys.stderr = buffer

    try:
        variables = {}
        exec(code,variables)
        response = {}
        for name,v in target_names.items():
            if name in variables:
                response[name] = variables[name]
    except Exception:
        return 1,traceback.format_exc(),{}

    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__

    return 0,buffer.getvalue(),response

def function_impl_format(prompt:str,func:Optional[Union[Callable,str]],
                             cls:Union[pydantic.BaseModel,str])->str:
    
    tool_choice_ser = serialize_function_to_json(func)    
    _cls = ""
    if isinstance(cls, str):
        _cls = cls
    else:
        _cls = cls.schema_json(ensure_ascii=False)

    example = '''{
  "name": "caculate_current_time",
  "description": "\n    计算当前时间\n    ",
  "parameters": {
    "type": "object",
    "properties": {}
  }
}'''    
    example_output_format='''
{"title": "CurrentTime", "description": "当前时间    ", "type": "object", "properties": {"time": {"title": "Time", "description": "开始时间.时间格式为 yyyy-MM-dd", "type": "string"}}, "required": ["time"]}
'''
    example_output = '''
from datetime import datetime

def caculate_current_time():
    # 获取当前日期和时间
    now = datetime.now()
    
    # 将日期和时间格式化为"yyyy-MM-dd"的形式
    time_str = now.strftime("%Y-%m-%d")
    
    return {"time": time_str}
'''
    
    msg = f''''你非常擅长 Python 语言。根据用户提供的一些信息以及问题，对提供了没有实现空函数函数进行实现。

示例：
你需要实现的函数的签名如下：

```json
{example}
```

生成的函数的返回值必须是 Json 格式，并且满足如下 OpenAPI 3.1. 规范：

```json
{example_output_format}
```

最后，你生成的函数的代码如下：

```python
{example_output}
```

现在，你需要实现函数的签名如下：

```json
{tool_choice_ser}
```

同时，你生成的函数的返回值必须是 Json 格式，并且满足如下 OpenAPI 3.1. 规范：

```json
{_cls}
```

用户的问题是：{prompt}

在满足上述提及的约束的情况下，请你实现这个函数。
注意：
1. 任何情况下都不要拆分成多段代码输出，请一次性生成完整的代码片段，确保代码的完整性
2. 回复的内容只有一个代码块，且代码块的语言为 Python
3. 不要演示如何调用你生成的函数的代码
'''
    return msg   



def function_calling_format(prompt:str,tools:List[Union[Callable,str]],tool_choice:Optional[Union[Callable,str]])->str:
    tool_serializes = []
    for v in tools:
        tool_serializes.append(serialize_function_to_json(v))

    force_prompt = ""
    if tool_choice is not None:
        tool_choice_ser = serialize_function_to_json(tool_choice)
        force_prompt = f''''
你必须使用如下的工具来解决用户的问题：        
```json
{tool_choice_ser}
```
'''  
   
    if tool_choice is None and len(tools) == 0:
        return prompt                   

    tools_str = "\n".join(tool_serializes)
    
    function_example = '''
{
  "name": "compute_date_range",
  "description": "\n    计算日期范围\n    ",
  "parameters": {
    "type": "object",
    "properties": {
      "count": {
        "type": "int",
        "description": "时间跨度，数值类型,如果用户说的是几天，几月啥的，比较模糊，务必使用默认值",
        "default": 3
      },
      "unit": {
        "enum": [
          "day",
          "week",
          "month",
          "year"
        ],
        "type": "str",
        "description": "",
        "default": "day"
      }
    }
  }
}
'''
    output_example = '''
{
  "id": "unique_id_1",
  "type": "function",
  "tool_calls": [
    {
      "function": {
        "name": "compute_date_range",
        "arguments": {
          "count": 3,
          "unit": "day"
        }
      }
    }
  ]
}
'''

    msg = f'''You are a helpful assistant with access to the following functions:

```json
{tools_str}
```

当用户的问题可以使用上面的一个或者多个函数解决时,你需要通过符合 OpenAPI 3.1 规范的 Json 格式告诉我你需要调用哪些函数。

下面Json文本描述了你需要返回的格式,它符合 OpenAPI 3.1 规范:

```json
{FUNCTION_CALLING_SCHEMA}
```

示例：

当你选择下面的函数时：

```
{function_example}
```

你应该使用如下的 Json 格式告诉我你需要调用这个函数：

```json
{output_example}
```

{force_prompt}

现在用户的问题是：{prompt}

请选择合适的一个或者多个函数按要求的 Json 格式返回给我。

注意：
1. 如果你无法使用上述函数解决用户的问题，请如实告诉我你没有办法回答。
''' 
    return msg  


def response_class_format(prompt:str,cls:Union[pydantic.BaseModel,str])->str:

    _cls = ""
    if isinstance(cls, str):
        _cls = cls
    else:
        _cls = cls.schema_json(ensure_ascii=False)    
        
    example='''
{"title": "Item", "description": "时间抽取的返回结果", "type": "object", "properties": {"time": {"title": "Time", "description": "时间信息,比如内容里会提到天， 月份，年等相关词汇", "type": "string"}, "other": {"title": "Other", "description": "除了时间以外的其他部分", "type": "string"}}, "required": ["time", "other"]}
'''
    example_output = '''{
  "time": "最近三个月",
  "other": "奔驰的销量趋势如何"
}'''
    msg = f'''当你回答用户问题的时候，你需要使用 Json 格式进行回复。

示例：

当你被要求按如下格式输出时,它符合 OpenAPI 3.1 规范：

```json
{example}
```
你的输出应该是这样的：

```json
{example_output}
```

现在用户的问题是：{prompt}

下面Json文本描述了你需要返回的格式,它符合 OpenAPI 3.1 规范:

```json
{_cls}
```

请根据自己生成的内容并以 Json 格式回复我。
''' 
    return msg

def response_class_format_after_chat(cls:Union[pydantic.BaseModel,str])->str:
 
    _cls = ""
    if isinstance(cls, str):
        _cls = cls
    else:
        _cls = cls.schema_json(ensure_ascii=False)
    example='''
{"title": "Item", "description": "时间抽取的返回结果", "type": "object", "properties": {"time": {"title": "Time", "description": "时间信息,比如内容里会提到天， 月份，年等相关词汇", "type": "string"}, "other": {"title": "Other", "description": "除了时间以外的其他部分", "type": "string"}}, "required": ["time", "other"]}
'''
    example_output = '''{
  "car": {
    "name": "奔驰"
  },
  "metric": {
    "name": "销量趋势"
  }'''    
    msg = f'''你需要以 Json 格式重新组织内容回复我。

示例：

当你被要求按如下格式输出时,它符合 OpenAPI 3.1 规范：

```json
{example}
```
你的输出应该是这样的：

```json
{example_output}
```
把你刚才回答我的内容重新做组织，以 Json 格式回复我

下面Json文本描述了你需要返回的格式,它符合 OpenAPI 3.1 规范:

```json
{_cls}
```
''' 
    return msg 

class BaseAbility(Enum):
    RESPONSE_WITH_CLASS = "RESPONSE_WITH_CLASS"
    FUNCTION_CALLING = "FUNCTION_CALLING"
    FUNCTION_IMPL = "FUNCTION_IMPL"
    OTHERS = "OTHERS"

def base_ability_format(prompt:Optional[str]=None,base_abilities: List[BaseAbility]=[BaseAbility.FUNCTION_CALLING,
                                                                                     BaseAbility.FUNCTION_IMPL,
                                                                                     BaseAbility.RESPONSE_WITH_CLASS                                                                                     
                                                                                     ])->str:
    base_abilities.extend([BaseAbility.OTHERS])

    RESPONSE_WITH_CLASS_example_0='''{"title": "Item", "description": "时间抽取的返回结果", "type": "object", "properties": {"time": {"title": "Time", "description": "时间信息,比如内容里会提到天， 月份，年等相关词汇", "type": "string"}, "other": {"title": "Other", "description": "除了时间以外的其他部分", "type": "string"}}, "required": ["time", "other"]}'''
    RESPONSE_WITH_CLASS_example_output_0 = '''{
    "time": "最近三个月",
    "other": "奔驰的销量趋势如何"
    }'''

    RESPONSE_WITH_CLASS_example='''{"title": "Info", "type": "object", "properties": {"car": {"title": "Car", "description": "车的信息", "allOf": [{"$ref": "#/definitions/Car"}]}, "metric": {"title": "Metric", "description": "计算的指标信息", "allOf": [{"$ref": "#/definitions/Metric"}]}}, "required": ["car", "metric"], "definitions": {"Car": {"title": "Car", "type": "object", "properties": {"name": {"title": "Name", "description": "品牌名称", "type": "string"}}, "required": ["name"]}, "Metric": {"title": "Metric", "type": "object", "properties": {"name": {"title": "Name", "description": "指标名称", "type": "string"}}, "required": ["name"]}}}'''
    RESPONSE_WITH_CLASS_example_output = '''{
  "car": {
    "name": "奔驰"
  },
  "metric": {
    "name": "销量趋势"
  }
}'''
    RESPONSE_WITH_CLASS_example='''{"title": "Info", "type": "object", "properties": {"car": {"title": "Car", "description": "车的信息", "allOf": [{"$ref": "#/definitions/Car"}]}, "metric": {"title": "Metric", "description": "计算的指标信息", "allOf": [{"$ref": "#/definitions/Metric"}]}}, "required": ["car", "metric"], "definitions": {"Car": {"title": "Car", "type": "object", "properties": {"name": {"title": "Name", "description": "品牌名称", "type": "string"}}, "required": ["name"]}, "Metric": {"title": "Metric", "type": "object", "properties": {"name": {"title": "Name", "description": "指标名称", "type": "string"}}, "required": ["name"]}}}'''
    RESPONSE_WITH_CLASS_example_output = '''{
  "car": {
    "name": "奔驰"
  },
  "metric": {
    "name": "销量趋势"
  }
}'''

    FUNCTION_CALLING_example = '''
{
  "name": "compute_date_range",
  "description": "\n    计算日期范围\n    ",
  "parameters": {
    "type": "object",
    "properties": {
      "count": {
        "type": "int",
        "description": "时间跨度，数值类型,如果用户说的是几天，几月啥的，比较模糊，务必使用默认值",
        "default": 3
      },
      "unit": {
        "enum": [
          "day",
          "week",
          "month",
          "year"
        ],
        "type": "str",
        "description": "",
        "default": "day"
      }
    }
  }
}
'''
    FUNCTION_CALLING_example_output = '''
{
  "id": "unique_id_1",
  "type": "function",
  "tool_calls": [
    {
      "function": {
        "name": "compute_date_range",
        "arguments": {
          "count": 3,
          "unit": "day"
        }
      }
    }
  ]
}
'''

    FUNCTION_IMPL_example = '''{
  "name": "caculate_current_time",
  "description": "\n    计算当前时间\n    ",
  "parameters": {
    "type": "object",
    "properties": {}
  }
}'''    
    FUNCTION_IMPL_example_output_schema='''
{"title": "CurrentTime", "description": "当前时间    ", "type": "object", "properties": {"time": {"title": "Time", "description": "开始时间.时间格式为 yyyy-MM-dd", "type": "string"}}, "required": ["time"]}
'''
    FUNCTION_IMPL_example_output = '''
from datetime import datetime

def caculate_current_time():
    # 获取当前日期和时间
    now = datetime.now()
    
    # 将日期和时间格式化为"yyyy-MM-dd"的形式
    time_str = now.strftime("%Y-%m-%d")
    
    return {"time": time_str}
'''
    m_response_class_str = "" if BaseAbility.RESPONSE_WITH_CLASS not in base_abilities else f'''
===================RESPONSE_WITH_CLASS===================

下面是一个根据用户的问题，并且结合 JSON Schema 生成对应的 JSON 数据的例子：

输入：

最近三个月奔驰的销量趋势如何？

JSON Schema：

```json
{RESPONSE_WITH_CLASS_example_0}
```

输出：

```json
{RESPONSE_WITH_CLASS_example_output_0}
```

下面生成的 Json 数据有有嵌套结构的例子：

输入：

最近三个月奔驰的销量趋势如何？

JSON Schema：

```json
{RESPONSE_WITH_CLASS_example}
```

输出：

```json
{RESPONSE_WITH_CLASS_example_output}
```

当用户提到 RESPONSE_WITH_CLASS 时，请回顾该能力。
'''
    m_function_calling_str = "" if BaseAbility.FUNCTION_CALLING not in base_abilities else f'''
===================FUNCTION_CALLING===================

用户会提供一个函数列表给你,你需要根据用户的问题，选择一个或者多个函数返回给用户。如果你无法使用上述函数解决用户的问题，请如实告诉我你没有办法回答。
下面假设你已经选择了一个函数作为输入，并且结合 JSON Schema 生成对应的 JSON 数据的例子：

输入：

```json
{FUNCTION_CALLING_example}
```

JSON Schema：

```json
{FUNCTION_CALLING_SCHEMA}
```

输出：

```json
{FUNCTION_CALLING_example_output}
```

当用户提到 FUNCTION_CALLING 时，请回顾该能力。
''' 
    m_function_impl_str = "" if BaseAbility.FUNCTION_IMPL not in base_abilities else f'''
===================FUNCTION_IMPL===================

你非常擅长 Python 语言。根据用户提供的一些信息以及问题，对用户提供的没有实现空函数函数进行实现。
下面假设用户提供了一个需要实现的函数的签名，你需要结合用户的问题，函数的签名，以及函数文档，生成对应的 Python 代码，函数的返回值
必须是 Json 格式，并且需要符合对应的 JSON Schema 规范。

下面提供了一个示例：

输入：

```json
{FUNCTION_IMPL_example}
```

JSON Schema：

```json
{FUNCTION_IMPL_example_output_schema}
```

输出：

```python
{FUNCTION_IMPL_example_output}
```

注意：
1. 任何情况下都不要拆分成多段代码输出，请一次性生成完整的代码片段，确保代码的完整性
2. 回复的内容只有一个代码块，且代码块的语言为 Python
3. 不要展示如何调用你生成的函数的代码
4. 不要展示你函数执行的结果

当用户提到 FUNCTION_IMPL 时，请回顾该能力。
'''
    msg = f'''下面是你具备的基础能力，当你回答用户问题的时候，随时回顾这些能力。

JSON 格式是一种轻量级的数据交换格式，JSON Schema 是基于 JSON 的一个描述 JSON 数据结构的元数据，可以用来描述 JSON 数据的结构和内容，以及定义 JSON 数据的合法值范围。
OpenAPI Specification (OAS) 使用 JSON Schema 来描述 Json 数据的结构和内容，你需要遵循 OpenAPI 3.1.0 版本的规范。

{m_response_class_str}

{m_function_calling_str}

{m_function_impl_str}

===================OTHERS===================
'''
    
    return msg


def sys_response_class_format(prompt:str,cls:Union[pydantic.BaseModel,str])->str:
    
    _cls = ""
    if isinstance(cls, str):
        _cls = cls
    else:
        _cls = cls.schema_json(ensure_ascii=False)

    msg = f'''
请使用 RESPONSE_WITH_CLASS 相关的能力，解决用户的问题。

输入：

{prompt}

JSON Schema：

```json
{_cls}
```

输出：
'''
    return msg

def sys_function_calling_format(prompt:str,tools:List[Union[Callable,str]],tool_choice:Optional[Union[Callable,str]])->str:
    tool_serializes = []
    for v in tools:
        tool_serializes.append(serialize_function_to_json(v))

    force_prompt = ""
    if tool_choice is not None:
        tool_choice_ser = serialize_function_to_json(tool_choice)
        force_prompt = f''''
你必须使用如下的工具来解决用户的问题：        
```json
{tool_choice_ser}
```
'''  
   
    if tool_choice is None and len(tools) == 0:
        return prompt                   

    tools_str = "\n".join(tool_serializes)
        
    msg = f'''
请使用 FUNCTION_CALLING 相关的能力，解决用户的问题。

你有如下的函数可以使用：

```json
{tools_str}
```
{force_prompt}

输入：

{prompt}

JSON Schema：

```json
{FUNCTION_CALLING_SCHEMA}
```

输出:
''' 
    return msg 

def sys_function_impl_format(prompt:str,func:Optional[Union[Callable,str]],
                             cls:Union[pydantic.BaseModel,str])->str:
    
    tool_choice_ser = serialize_function_to_json(func)    
    _cls = ""
    if isinstance(cls, str):
        _cls = cls
    else:
        _cls = cls.schema_json(ensure_ascii=False)

    
    msg = f''''请使用 FUNCTION_IMPL 相关的能力，解决用户的问题。
根据用户提供的一些信息以及函数签名，对函数进行实现。

用户问题： {prompt}

输入：

```json
{tool_choice_ser}
```

JSON Schema：

```json
{_cls}
```

输出:
'''
    return msg  

def format_prompt(func,**kargs): 
    from langchain import PromptTemplate
    doc = func.__doc__       
    lines = doc.splitlines()
    # get the first line to get the whitespace prefix
    first_non_empty_line = next(line for line in lines if line.strip())
    prefix_whitespace_length = len(first_non_empty_line) - len(first_non_empty_line.lstrip())    
    prompt = "\n".join([line[prefix_whitespace_length:] for line in lines])
    tpl = PromptTemplate.from_template(prompt)
    return tpl.format(**kargs)

def format_prompt_jinja2(func,**kargs):
    from jinja2 import Template
    doc = func.__doc__       
    lines = doc.splitlines()
    # get the first line to get the whitespace prefix
    first_non_empty_line = next(line for line in lines if line.strip())
    prefix_whitespace_length = len(first_non_empty_line) - len(first_non_empty_line.lstrip())    
    prompt = "\n".join([line[prefix_whitespace_length:] for line in lines])
    tpl = Template(prompt)
    return tpl.render(kargs)

def random_uuid() -> str:
    return str(uuid.uuid4().hex)


__all__ = ["VLLMStreamServer", "BlockVLLMStreamServer","StreamOutputs","SingleOutput","SingleOutputMeta","BlockBinaryStreamServer"]




from typing import List,Optional,Dict
from byzerllm.utils.client.types import (
    Templates,Template,Role,LLMHistoryItem,
    LLMRequest,
    LLMFunctionCallResponse,
    LLMClassResponse,InferBackend,EventName,EventCallbackResult,EventCallback,LLMResponse,FintuneRequestExtra,
    FintuneRequest,ExecuteCodeResponse,LLMMetadata
)
from byzerllm.utils.client.byzerllm_client import ByzerLLM

def default_chat_wrapper(llm:ByzerLLM,conversations: Optional[List[Dict]] = None,llm_config={}):
    return llm.chat_oai(conversations=conversations,llm_config=llm_config)

__all__ = [
    "ByzerLLM","default_chat_wrapper","Templates","Template","Role","LLMHistoryItem",
    "LLMRequest",
    "LLMFunctionCallResponse",
    "LLMClassResponse","InferBackend","EventName","EventCallbackResult","EventCallback","LLMResponse","FintuneRequestExtra",
    "FintuneRequest","ExecuteCodeResponse"]





"""
Much of this code is adapted from Andrej Karpathy's NanoGPT
(https://github.com/karpathy/nanoGPT)
"""
from dataclasses import dataclass
import math

import torch
import torch.nn as nn
from torch.nn import functional as F

from .model import GPT, GPTConfig, MLP


class NonCausalSelfAttention(nn.Module):
    def __init__(self, config):
        super().__init__()
        assert config.n_embd % config.n_head == 0
        # key, query, value projections for all heads, but in a batch
        self.c_attn = nn.Linear(config.n_embd, 3 * config.n_embd, bias=config.bias)
        # output projection
        self.c_proj = nn.Linear(config.n_embd, config.n_embd, bias=config.bias)
        # regularization
        self.attn_dropout = nn.Dropout(config.dropout)
        self.resid_dropout = nn.Dropout(config.dropout)
        self.n_head = config.n_head
        self.n_embd = config.n_embd
        self.dropout = config.dropout
        # flash attention make GPU go brrrrr but support is only in PyTorch nightly and still a bit scary
        self.flash = (
            hasattr(torch.nn.functional, "scaled_dot_product_attention") and self.dropout == 0.0
        )

    def forward(self, x):
        B, T, C = x.size()  # batch size, sequence length, embedding dimensionality (n_embd)

        # calculate query, key, values for all heads in batch and move head forward to be the batch dim
        q, k, v = self.c_attn(x).split(self.n_embd, dim=2)
        k = k.view(B, T, self.n_head, C // self.n_head).transpose(1, 2)  # (B, nh, T, hs)
        q = q.view(B, T, self.n_head, C // self.n_head).transpose(1, 2)  # (B, nh, T, hs)
        v = v.view(B, T, self.n_head, C // self.n_head).transpose(1, 2)  # (B, nh, T, hs)

        # causal self-attention; Self-attend: (B, nh, T, hs) x (B, nh, hs, T) -> (B, nh, T, T)
        if self.flash:
            # efficient attention using Flash Attention CUDA kernels
            y = torch.nn.functional.scaled_dot_product_attention(
                q, k, v, attn_mask=None, dropout_p=self.dropout, is_causal=False
            )
        else:
            # manual implementation of attention
            att = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.size(-1)))
            att = F.softmax(att, dim=-1)
            att = self.attn_dropout(att)
            y = att @ v  # (B, nh, T, T) x (B, nh, T, hs) -> (B, nh, T, hs)
        y = (
            y.transpose(1, 2).contiguous().view(B, T, C)
        )  # re-assemble all head outputs side by side

        # output projection
        y = self.resid_dropout(self.c_proj(y))
        return y


class FineBlock(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.ln_1 = nn.LayerNorm(config.n_embd)
        self.attn = NonCausalSelfAttention(config)
        self.ln_2 = nn.LayerNorm(config.n_embd)
        self.mlp = MLP(config)

    def forward(self, x):
        x = x + self.attn(self.ln_1(x))
        x = x + self.mlp(self.ln_2(x))
        return x


class FineGPT(GPT):
    def __init__(self, config):
        super().__init__(config)
        del self.lm_head
        self.config = config
        self.n_codes_total = config.n_codes_total
        self.transformer = nn.ModuleDict(
            dict(
                wtes=nn.ModuleList(
                    [
                        nn.Embedding(config.input_vocab_size, config.n_embd)
                        for _ in range(config.n_codes_total)
                    ]
                ),
                wpe=nn.Embedding(config.block_size, config.n_embd),
                drop=nn.Dropout(config.dropout),
                h=nn.ModuleList([FineBlock(config) for _ in range(config.n_layer)]),
                ln_f=nn.LayerNorm(config.n_embd),
            )
        )
        self.lm_heads = nn.ModuleList(
            [
                nn.Linear(config.n_embd, config.output_vocab_size, bias=False)
                for _ in range(config.n_codes_given, self.n_codes_total)
            ]
        )
        for i in range(self.n_codes_total - config.n_codes_given):
            self.transformer.wtes[i + 1].weight = self.lm_heads[i].weight

    def forward(self, pred_idx, idx):
        device = idx.device
        b, t, codes = idx.size()
        assert (
            t <= self.config.block_size
        ), f"Cannot forward sequence of length {t}, block size is only {self.config.block_size}"
        assert pred_idx > 0, "cannot predict 0th codebook"
        assert codes == self.n_codes_total, (b, t, codes)
        pos = torch.arange(0, t, dtype=torch.long, device=device).unsqueeze(0)  # shape (1, t)

        # forward the GPT model itself
        tok_embs = [
            wte(idx[:, :, i]).unsqueeze(-1) for i, wte in enumerate(self.transformer.wtes)
        ]  # token embeddings of shape (b, t, n_embd)
        tok_emb = torch.cat(tok_embs, dim=-1)
        pos_emb = self.transformer.wpe(pos)  # position embeddings of shape (1, t, n_embd)
        x = tok_emb[:, :, :, : pred_idx + 1].sum(dim=-1)
        x = self.transformer.drop(x + pos_emb)
        for block in self.transformer.h:
            x = block(x)
        x = self.transformer.ln_f(x)
        logits = self.lm_heads[pred_idx - self.config.n_codes_given](x)
        return logits

    def get_num_params(self, non_embedding=True):
        """
        Return the number of parameters in the model.
        For non-embedding count (default), the position embeddings get subtracted.
        The token embeddings would too, except due to the parameter sharing these
        params are actually used as weights in the final layer, so we include them.
        """
        n_params = sum(p.numel() for p in self.parameters())
        if non_embedding:
            for wte in self.transformer.wtes:
                n_params -= wte.weight.numel()
            n_params -= self.transformer.wpe.weight.numel()
        return n_params


@dataclass
class FineGPTConfig(GPTConfig):
    n_codes_total: int = 8
    n_codes_given: int = 1



import diffusers

# TODO: add rest Auto111 samplers
# Commented schedulers are still unavailable in diffusers 0.16.1
SCHEDULERS = {
    "unipc": diffusers.schedulers.UniPCMultistepScheduler,
    "euler_a": diffusers.schedulers.EulerAncestralDiscreteScheduler,
    "euler": diffusers.schedulers.EulerDiscreteScheduler,
    "ddim": diffusers.schedulers.DDIMScheduler,
    "ddpm": diffusers.schedulers.DDPMScheduler,
    "deis": diffusers.schedulers.DEISMultistepScheduler,
    "dpm2": diffusers.schedulers.KDPM2DiscreteScheduler,
    "dpm2-a": diffusers.schedulers.KDPM2AncestralDiscreteScheduler,
    "dpm++_2s": diffusers.schedulers.DPMSolverSinglestepScheduler,
    "dpm++_2m": diffusers.schedulers.DPMSolverMultistepScheduler,
    "dpm++_2m_karras": diffusers.schedulers.DPMSolverMultistepScheduler,
    "heun": diffusers.schedulers.HeunDiscreteScheduler,
    "heun_karras": diffusers.schedulers.HeunDiscreteScheduler,
    "lms": diffusers.schedulers.LMSDiscreteScheduler,
    "pndm": diffusers.schedulers.PNDMScheduler,
}


def parser_schedulers_config(scheduler_id: str):
    """
    add extra config parameter to scheduler
    """
    kwargs = {}
    if "karras" in scheduler_id:
        kwargs["use_karras_sigmas"] = True
    return kwargs



from ..conversable_agent import ConversableAgent
from ..agent import Agent
from .. import get_agent_name,run_agent_func,ChatResponse
import ray
from ray.util.client.common import ClientActorHandle, ClientObjectRef
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from typing import Callable, Dict, Optional, Union
import time
import sys
import io
import traceback
import json
import os
from byzerllm.utils.retrieval import ByzerRetrieval
from byzerllm.utils.client import ByzerLLM,code_utils,message_utils



class CodeSandbox:
    def __init__(self,file_path:str,file_ref) -> None:        
        self.file_ref = file_ref
        self.file_path = file_path
        self.session_variables = {}
        if self.file_ref:
            if isinstance(self.file_ref,ClientObjectRef):
                content = ray.get(self.file_ref)
            else:
                content = self.file_ref   

            # check parent directory of self.file_path exists
            parent_dir = os.path.dirname(self.file_path)
            if not os.path.exists(parent_dir):
                os.makedirs(parent_dir)     
                     
            with open(self.file_path, "wb") as f:
                f.write(content)
                
    def set_value(self,name:str,value:str): 
        self.session_variables[name] = value
        return self

    def get_value(self,name:str):
        if name not in self.session_variables:
            return None
        return self.session_variables[name]

    def get_file_path(self):
        return self.file_path        

    def execute_code(self,code)->Tuple[int, str, str]:
        return code_utils.execute_code(
                code = code,
                timeout=30*60,
                filename=None,
                work_dir=None,
                use_docker=False,
                lang="python"        
                ) 
    
    def exec_capture_output(self,code: str,target_names:Dict[str,Any]={}) -> Tuple[int,str,Any]:
        buffer = io.StringIO()
        sys.stdout = buffer
        sys.stderr = buffer

        try:
            variables = {}
            exec(code,variables)
            response = {}
            for name,v in target_names.items():
                if name in variables:
                    response[name] = variables[name]
        except Exception:
            return 1,traceback.format_exc(),{}

        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

        return 0,buffer.getvalue(),response
    
class PythonSandboxAgent(ConversableAgent):        

    def __init__(
        self,
        name: str,
        llm: ByzerLLM,
        retrieval: ByzerRetrieval,   
        chat_name:str,
        owner:str,     
        system_message: Optional[str],        
        is_termination_msg: Optional[Callable[[Dict], bool]] = None,
        max_consecutive_auto_reply: Optional[int] = None,
        human_input_mode: Optional[str] = "NEVER",
        code_execution_config: Optional[Union[Dict, bool]] = {},
        **kwargs,
    ):
        super().__init__(
            name,
            llm,retrieval,
            system_message,
            is_termination_msg,
            max_consecutive_auto_reply,
            human_input_mode,
            code_execution_config=code_execution_config,            
            **kwargs,
        )
        self.sandboxes = {}
        self.lasted_updated = {}
        
        ## Restore the reply function list
        self._reply_func_list = []        

        ## Register the reply functions                
        self.register_reply([Agent, ClientActorHandle,str], PythonSandboxAgent.generate_execute_code_reply) 
        self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.check_termination_and_human_reply)             
        

    def generate_execute_code_reply(
        self,
        raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Union[ClientActorHandle,Agent,str]] = None,
        config: Optional[Any] = None,
    ) -> Tuple[bool, Union[str, Dict, None,ChatResponse]]:
                
        code_execution_config = config if config is not None else self._code_execution_config
        
        if code_execution_config is False:            
            return False, None
        
        if messages is None:
            messages = self._messages[get_agent_name(sender)]
        
        last_n_messages = code_execution_config.pop("last_n_messages", 1)                

        for i in range(min(len(messages), last_n_messages)):
            message = messages[-(i + 1)]
            if not message["content"]:
                continue            
            
            code_blocks = code_utils.extract_code(message["content"])
            if len(code_blocks) == 1 and code_blocks[0][0] == "unknown":
                continue

            # found code blocks, execute code and push "last_n_messages" back
            #  combine all code blocks into one code block
            codes = [code_block[1] for code_block in code_blocks if code_block[0] == "python"]
            code_str = "\n".join(codes)
            
            file_path = None
            file_ref = None

            if "metadata" not in message:
                message["metadata"] = {}
            
            if "file_path" in message["metadata"]:
                file_path = message["metadata"]["file_path"]
                file_ref = message["metadata"]["file_ref"]                
            
            target_names = {}
            if "target_names" in message["metadata"]:
                target_names = message["metadata"]["target_names"]

            sandbox = self.get_or_create_sandbox(get_agent_name(sender)+"_sandbox",file_path,file_ref,0,0)            
            exitcode, output,response = ray.get(sandbox.exec_capture_output.remote(code_str,target_names))
            code_execution_config["last_n_messages"] = last_n_messages
            exitcode2str = "execution succeeded" if exitcode == 0 else "execution failed"
            
            return True, {
                "content":f"exitcode: {exitcode} ({exitcode2str})\nCode output: {output}",
                 "metadata":{
                     "execute_result":ChatResponse(status=exitcode,
                                      output=f"exitcode: {exitcode} ({exitcode2str})\nCode output: {output}",
                                      code=code_str,
                                      prompt=message,
                                      variables=response,                                      
                                      ),
                      "error_count": message_utils.get_error_count(message),
                 }
            }

        print("No code block found in the last {} messages.".format(last_n_messages),flush=True)
        code_execution_config["last_n_messages"] = last_n_messages

        return True, None            

    def check_sandbox_timeout(self,timeout:int=60*60): 
        remove_names = []
        for name in self.lasted_updated:
            if time.time() - self.lasted_updated[name] > timeout:
                remove_names.append(name)
        for name in remove_names:
            del self.sandboxes[name]
            del self.lasted_updated[name]        

    def check_sandbox_exists(self,name:str)->bool:
        return name in self.sandboxes

    def get_sandbox(self,name:str):                
        self.check_sandbox_timeout()        
        return self.sandboxes[name]
    
    def force_clear(self):
        self.sandboxes = {}
        self.lasted_updated = {}

    def get_or_create_sandbox(self,name:str,
                              file_path:str,file_ref:str,
                              num_gpus:int,num_cpus:int):
        self.lasted_updated[name] = time.time()
        self.check_sandbox_timeout()
        if name in self.sandboxes:            
            return self.sandboxes[name]
        
        try :
            sandbox = ray.get_actor(name)
            return sandbox
        except ValueError:
            pass
        
        sandbox = ray.remote(CodeSandbox).options(
                name=name,                                             
                num_cpus=num_cpus,
                num_gpus=num_gpus
            ).remote(file_path,file_ref)
        self.sandboxes[name] = sandbox
        return sandbox    


import _thread as thread
import base64
import datetime
import hashlib
import hmac
import json
from urllib.parse import urlparse
import ssl
from datetime import datetime
from time import mktime
from urllib.parse import urlencode
from wsgiref.handlers import format_date_time
from typing import List, Tuple,Dict,Any
from byzerllm.utils.langutil import asyncfy_with_semaphore
import queue

import websocket

reponse_queue = queue.Queue()

class SparkDeskAPIParams(object):
    # 初始化
    def __init__(self, APPID, APIKey, APISecret, gpt_url, DOMAIN):
        self.APPID = APPID
        self.APIKey = APIKey
        self.APISecret = APISecret
        self.host = urlparse(gpt_url).netloc
        self.path = urlparse(gpt_url).path
        self.gpt_url = gpt_url
        self.DOMAIN = DOMAIN

        # 生成url
    def create_url(self):
        # 生成RFC1123格式的时间戳
        now = datetime.now()
        date = format_date_time(mktime(now.timetuple()))

        # 拼接字符串
        signature_origin = "host: " + self.host + "\n"
        signature_origin += "date: " + date + "\n"
        signature_origin += "GET " + self.path + " HTTP/1.1"

        # 进行hmac-sha256进行加密
        signature_sha = hmac.new(self.APISecret.encode('utf-8'), signature_origin.encode('utf-8'),
                                 digestmod=hashlib.sha256).digest()

        signature_sha_base64 = base64.b64encode(signature_sha).decode(encoding='utf-8')

        authorization_origin = f'api_key="{self.APIKey}", algorithm="hmac-sha256", headers="host date request-line", signature="{signature_sha_base64}"'

        authorization = base64.b64encode(authorization_origin.encode('utf-8')).decode(encoding='utf-8')

        # 将请求的鉴权参数组合为字典
        v = {
            "authorization": authorization,
            "date": date,
            "host": self.host
        }
        # 拼接鉴权参数，生成url
        url = self.gpt_url + '?' + urlencode(v)
        # 此处打印出建立连接时候的url,参考本demo的时候可取消上方打印的注释，比对相同参数时生成的url与自己代码生成的url是否一致
        return url


class CustomSaasAPI:

    def __init__(self, infer_params: Dict[str, str]) -> None:
        required_params = [ "saas.appid", "saas.api_key", "saas.api_secret"]
        for param in required_params:
            if list(infer_params.keys()).count(param) < 1:
                raise ValueError("The parameter %s is a required field, please configure it"% param)
        for value in self.get_value(infer_params,required_params):
            if value is None or value == "":
                raise ValueError("The mandatory model parameters cannot be empty.")
        self.appid: str = infer_params["saas.appid"]
        self.api_key: str = infer_params["saas.api_key"]
        self.api_secret: str = infer_params["saas.api_secret"]
        self.gpt_url: str = infer_params.get("saas.gpt_url","wss://spark-api.xf-yun.com/v3.1/chat")
        self.domain: str = infer_params.get("saas.domain","generalv3")
        self.config = SparkDeskAPIParams(self.appid, self.api_key, self.api_secret, self.gpt_url, self.domain)
        self.timeout = int(infer_params.get("saas.timeout",30))
        self.debug = infer_params.get("saas.debug",False)

    @staticmethod
    def on_error(ws, error):
        pass


    @staticmethod
    def on_close(ws,a,b):
        pass


    @staticmethod
    def on_open(ws):
        thread.start_new_thread(CustomSaasAPI.run, (ws,))

    @staticmethod
    def run(ws, *args):
        # 8192
        data = {
            "header": {
                "app_id": ws.appid,
                "uid": "1234"
            },
            "parameter": {
                "chat": {
                    "domain": ws.domain,
                    "random_threshold": ws.temperature,
                    "max_tokens": ws.max_length,
                    "auditing": "default"
                }
            },
            "payload": {
                "message": {
                    "text": ws.question
                }
            }
        }
        data = json.dumps(data)        
        ws.send(data)


    @staticmethod
    def on_message(ws, message):
        data = json.loads(message)        
        code = data['header']['code']
        if code != 0:
            reponse_queue.put(f'请求错误: {code}, {data}')
            reponse_queue.put(None)
            ws.close()
        else:
            choices = data["payload"]["choices"]
            status = choices["status"]
            content = choices["text"][0]["content"]
            reponse_queue.put(content)
            if status == 2:
                reponse_queue.put(None)
                ws.close()


    # saas/proprietary
    def get_meta(self):
        return [{
            "model_deploy_type": "saas",
            "backend":"saas"
        }]
    
    async def async_get_meta(self):
        return await asyncfy_with_semaphore(self.get_meta)()

    def get_value(self,infer_params: Dict[str, str],keys_to_get):
        values = []
        for key in keys_to_get:
            if key in infer_params.keys():
                values.append(infer_params[key])
        return values
    
    async def async_stream_chat(self,tokenizer,ins:str, his:List[Dict[str,Any]]=[],
                    max_length:int=4096,
                    top_p:float=0.7,
                    temperature:float=0.9):
        return await asyncfy_with_semaphore(self.stream_chat)(tokenizer,ins,his,max_length,top_p,temperature)

    def stream_chat(self,tokenizer,ins:str, his:List[Dict[str,Any]]=[],
                    max_length:int=4096,
                    top_p:float=0.7,
                    temperature:float=0.9):

        q = his + [{"role": "user", "content": ins}]
        websocket.enableTrace(self.debug)
        wsUrl = self.config.create_url()
        ws = websocket.WebSocketApp(wsUrl,
                                    on_message=CustomSaasAPI.on_message,
                                    on_error=CustomSaasAPI.on_error,
                                    on_close=CustomSaasAPI.on_close,
                                    on_open=CustomSaasAPI.on_open)
        ws.appid = self.config.APPID
        ws.domain = self.config.DOMAIN
        ws.question = q
        ws.max_length = max_length
        ws.top_p = top_p
        ws.temperature = temperature
        ws.run_forever(sslopt={"cert_reqs": ssl.CERT_NONE})

        result = []

        t  = reponse_queue.get(timeout=self.timeout)
        while t is not None:
            result.append(t)
            t  = reponse_queue.get(timeout=self.timeout)       
         
        return [("".join(result),"")]


from typing import List,Optional
import os
import subprocess
import re
from loguru import logger
from byzerllm.apps.byzer_storage import env

def _check_java_version(java_home:str):
    try:
        output = subprocess.check_output([f"{java_home}/bin/java", "-version"], stderr=subprocess.STDOUT, universal_newlines=True)
        version_line = output.splitlines()[0]
        version_match = re.search(r'version "(\d+)', version_line)
        if version_match:
            version = version_match.group(1)
            version_parts = version.split(".")
            major_version = int(version_parts[0])
            print(major_version)
            if major_version < 21:
                raise ValueError(f"Java version {version} is not supported. JDK 21 or higher is required.")
        else:
            raise ValueError("Could not determine Java version.")
    except (subprocess.CalledProcessError, ValueError) as e:
        raise ValueError(f"Error checking Java version: {str(e)}")



def connect_cluster(address:str="auto",java_home:Optional[str]=None,
                code_search_path:Optional[List[str]]=None):
    import ray
    job_config = None
    env_vars = {}
    
    java_home=java_home if java_home else os.environ.get("JAVA_HOME")
    
    v = env.detect_env() 
    if v.java_home and v.java_version == "21": 
        logger.info(f"JDK 21 will be used ({v.java_home})...")    
        java_home = v.java_home        
    
    if java_home:            
        path = os.environ.get("PATH")    
        env_vars = {"JAVA_HOME": java_home,
                    "PATH":f'''{os.path.join(java_home,"bin")}:{path}'''}        
        if code_search_path:
            if java_home:
                _check_java_version(java_home)
            job_config = ray.job_config.JobConfig(code_search_path=code_search_path,
                                                            runtime_env={"env_vars": env_vars})
    if not java_home and code_search_path:
       logger.warning("code_search_path is ignored because JAVA_HOME is not set")         

    ray.init(address=address,namespace="default",ignore_reinit_error=True,
                    job_config=job_config)
    return env_vars 



from byzerllm.utils.retrieval import ByzerRetrieval
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from ....utils.client import ByzerLLM
from ....utils import generate_str_md5
from byzerllm.utils.retrieval import ByzerRetrieval
import time
from byzerllm.utils.client import LLMHistoryItem,LLMRequest
from byzerllm.utils.retrieval import TableSettings,SearchQuery
import uuid
import json
from langchain import PromptTemplate
from langchain.text_splitter import RecursiveCharacterTextSplitter,Document
try:
    from termcolor import colored
except ImportError:

    def colored(x, *args, **kwargs):
        return x

import jieba  

class SimpleRetrievalClient:
    def __init__(self,llm:ByzerLLM, retrieval: ByzerRetrieval,retrieval_cluster:str,retrieval_db:str,max_output_length=10000):
        self.retrieval_cluster = retrieval_cluster
        self.retrieval_db = retrieval_db
        self.max_output_length = max_output_length
        self.llm = llm
        self.retrieval = retrieval                
        if self.llm.default_emb_model_name is None:
            raise Exception(f'''
emb model does not exist. Try to use `llm.setup_default_emb_model_name` to set the default emb model name.
''')        

        # create the retrieval database/table if not exists
        if self.retrieval and not self.retrieval.check_table_exists(self.retrieval_cluster,self.retrieval_db,"text_content"):
           self.retrieval.create_table(self.retrieval_cluster,tableSettings=TableSettings(
                database=self.retrieval_db,
                table="text_content",schema='''st(
field(_id,string),
field(owner,string),
field(title,string,analyze),
field(content,string,analyze),
field(url,string),
field(raw_content,string),
field(auth_tag,string,analyze),
field(title_vector,array(float)),
field(content_vector,array(float)),
field(created_time,long,sort)
)''',
                location=f"/tmp/{self.retrieval_cluster}",num_shards=1 
           ))

           self.retrieval.create_table(self.retrieval_cluster,tableSettings=TableSettings(
                database=self.retrieval_db,
                table="text_content_chunk",schema='''st(
field(_id,string),
field(doc_id,string),
field(owner,string),
field(chunk,string,analyze),
field(raw_chunk,string),
field(chunk_vector,array(float)),
field(created_time,long,sort)
)''',
                location=f"/tmp/{self.retrieval_cluster}",num_shards=1                
           )) 
           if not self.retrieval.check_table_exists(self.retrieval_cluster,self.retrieval_db,"user_memory"):
                self.retrieval.create_table(self.retrieval_cluster,tableSettings=TableSettings(
                        database=self.retrieval_db,
                        table="user_memory",schema='''st(
        field(_id,string),
        field(chat_name,string),
        field(role,string),
        field(owner,string),
        field(content,string,analyze),
        field(raw_content,string),
        field(auth_tag,string,analyze),
        field(created_time,long,sort),
        field(chat_name_vector,array(float)),
        field(content_vector,array(float))
        )
        ''',
                        location=f"/tmp/{self.retrieval_cluster}",num_shards=1
                ))  

    def save_conversation(self,owner:str,chat_name:str,role:str,content:str):
        if not self.retrieval:
            raise Exception("retrieval is not setup")                                

        if chat_name is None:
            chat_name = content[0:10]   

        if len(content) > self.max_output_length:
            raise Exception(f"The response content length {len(content)} is larger than max_output_length {self.max_output_length}")

        data = [{"_id":str(uuid.uuid4()),
                "chat_name":chat_name,
                "role":role,
                "owner":owner,
                "content":self.search_tokenize(content),
                "raw_content":content,
                "auth_tag":"",
                "created_time":int(time.time()*1000),
                "chat_name_vector":self.emb(chat_name),
                "content_vector":self.emb(content)}]    

        self.retrieval.build_from_dicts(self.retrieval_cluster,self.retrieval_db,"user_memory",data)

    def get_conversations(self,owner:str, chat_name:str,limit=1000)->List[Dict[str,Any]]:
        docs = self.retrieval.filter(self.retrieval_cluster,
                        [SearchQuery(self.retrieval_db,"user_memory",
                                     filters={"and":[self._owner_filter(owner),{"field":"chat_name","value":chat_name}]},
                                     sorts=[{"created_time":"desc"}],
                                    keyword=None,fields=["chat_name"],
                                    vector=[],vectorField=None,
                                    limit=limit)])
        sorted_docs = sorted(docs[0:limit],key=lambda x:x["created_time"],reverse=False)
        return sorted_docs
    
    def get_conversations_as_history(self,owner:str,chat_name:str,limit=1000)->List[LLMHistoryItem]:
        chat_history = self.get_conversations(owner,chat_name,limit=limit)        
        chat_history = [LLMHistoryItem(item["role"],item["raw_content"]) for item in chat_history]
        return chat_history    


    def save_text_content(self,owner:str,title:str,content:str,url:str,auth_tag:str="",auto_chunking:bool=True):

        if not self.retrieval:
            raise Exception("retrieval is not setup")

                        
        text_content = [{"_id":generate_str_md5(content),
            "title":self.search_tokenize(title),
            "content":self.search_tokenize(content[0:10000]),
            "owner":owner,
            "raw_content":content[0:10000],
            "url":url,
            "auth_tag":self.search_tokenize(auth_tag),
            "title_vector":self.emb(title),
            "content_vector":self.emb(content[0:2048]),
            "created_time":int(time.time()*1000),
            }]
        self.retrieval.build_from_dicts(self.retrieval_cluster,self.retrieval_db,"text_content",text_content)
        
        if auto_chunking:            
            content_chunks= self.split_text_into_chunks(content)
                        
            text_content_chunks = [{"_id":f'''{text_content[0]["_id"]}_{i}''',
                "doc_id":text_content[0]["_id"],
                "owner":owner,
                "chunk":self.search_tokenize(item),
                "raw_chunk":item,
                "chunk_vector":self.emb(item),
                "created_time":int(time.time()*1000),
                } for i,item in enumerate(content_chunks)]
            
            self.retrieval.build_from_dicts(self.retrieval_cluster,self.retrieval_db,"text_content_chunk",text_content_chunks)    

    
    def _owner_filter(self,owner:str):
        return {"field":"owner","value":owner}
            
    def search_content_chunks(self,q:str,owner:str,limit:int=4,return_json:bool=True):   
        docs = self.retrieval.search(self.retrieval_cluster,
                            [SearchQuery(self.retrieval_db,"text_content_chunk",
                                         filters={"and":[self._owner_filter(owner)]},
                                        keyword=self.search_tokenize(q),fields=["content"],
                                        vector=self.emb(q),vectorField="content_vector",
                                        limit=limit)])

        if return_json:
            context = json.dumps([{"content":x["raw_chunk"]} for x in docs],ensure_ascii=False,indent=4)    
            return context 
        else:
            return docs

    def search_content(self,q:str,owner:str,url:str,auth_tag:str=None,limit:int=4,return_json:bool=True): 
        filters = [self._owner_filter(owner)]
        
        if auth_tag:
            filters.append({"field":"auth_tag","value":self.search_tokenize(auth_tag)})
        
        if url:
            filters.append({"field":"url","value":url})    

        if q:
            keyword = self.search_tokenize(q)
            vector = self.emb(q)
            vectorField = "content_vector"
            fields = ["content"]
        else:
            keyword = None
            vector = []
            vectorField = None
            fields = []

        docs = self.retrieval.search(self.retrieval_cluster,
                            [SearchQuery(self.retrieval_db,"text_content",
                                         filters={"and":filters},
                                        keyword=keyword,fields=fields,
                                        vector=vector,vectorField=vectorField,
                                        limit=limit)])

        if return_json:
            context = json.dumps([{"content":x["raw_content"]} for x in docs],ensure_ascii=False,indent=4)    
            return context 
        else:
            return docs  
        
    def get_doc(self,doc_id:str,owner:str):
        docs = self.retrieval.search(self.retrieval_cluster,
                            [SearchQuery(self.retrieval_db,"text_content",
                                         filters={"and":[self._owner_filter(owner)]},
                                        keyword=doc_id,fields=["_id"],
                                        vector=[],vectorField=None,
                                        limit=1)])
        return docs[0] if docs else None
    
    def get_doc_by_url(self,url:str,owner:str):
        docs = self.retrieval.search(self.retrieval_cluster,
                            [SearchQuery(self.retrieval_db,"text_content",
                                         filters={"and":[self._owner_filter(owner)]},
                                        keyword=url,fields=["url"],
                                        vector=[],vectorField=None,
                                        limit=1)])
        return docs[0] if docs else None
                
        
    def search_memory(self,chat_name:str, owner:str, q:str,limit:int=4,return_json:bool=True):
        docs = self.retrieval.search(self.retrieval_cluster,
                        [SearchQuery(self.retrieval_db,"user_memory",
                                    filters={"and":[self._owner_filter(owner=owner),{"field":"chat_name","value":chat_name}]},
                                    sorts =[{"created_time":"desc"}],
                                    keyword=self.search_tokenize(q),fields=["content"],
                                    vector=self.emb(q),vectorField="content_vector",
                                    limit=1000)])
        docs = [doc for doc in docs if doc["role"] == "user" and doc["chat_name"] == chat_name]
        if return_json:
            context = json.dumps([{"content":x["raw_chunk"]} for x in docs[0:limit]],ensure_ascii=False,indent=4)    
            return context 
        else:
            return docs[0:limit]    

    def emb(self,s:str):        
        return self.llm.emb(self.llm.default_emb_model_name,LLMRequest(instruction=s))[0].output[0:1024] 


    def split_text_into_chunks(self,s:str):
        # self.llm.apply_sql_func(
        #     '''select llm_split(value,array(",","。","\n"),1600) as value ''',[{"value":content}],
        #     url=self.byzer_engine_url
        #     )["value"]
        text_splitter = RecursiveCharacterTextSplitter(chunk_size=1600, chunk_overlap=200)
        split_docs = text_splitter.split_documents([Document(page_content=s)])         
        return [s.page_content for s in split_docs] 

    
    def search_tokenize(self,s:str):
        seg_list = jieba.cut(s, cut_all=False)
        # return self.llm.apply_sql_func("select mkString(' ',parse(value)) as value",[
        # {"value":s}],url=self.byzer_engine_url)["value"]
        return " ".join(seg_list) 
    
               
    




from ..conversable_agent import ConversableAgent
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from ....utils.client import ByzerLLM
from byzerllm.utils.retrieval import ByzerRetrieval
from ..agent import Agent
import ray
from ray.util.client.common import ClientActorHandle, ClientObjectRef
from .. import get_agent_name,run_agent_func,ChatResponse
from byzerllm.apps.llama_index.simple_retrieval import SimpleRetrieval
import uuid
import json
from langchain import PromptTemplate
from byzerllm.apps.llama_index import get_service_context,get_storage_context
from llama_index import SimpleDirectoryReader
from llama_index.node_parser import SentenceSplitter
from llama_index import SimpleDirectoryReader, VectorStoreIndex, ServiceContext
from llama_index.node_parser import SentenceSplitter,SentenceWindowNodeParser
from llama_index import (  
    Document,  
    get_response_synthesizer,
)
from llama_index.indices.document_summary import DocumentSummaryIndex
import pydantic

try:
    from termcolor import colored
except ImportError:

    def colored(x, *args, **kwargs):
        return x

import jieba  

class AgentData(pydantic.BaseModel):
    path:str = pydantic.Field(...,description="用户提及的路径全名")
    namespace:str = pydantic.Field(...,description="用户提及的命名空间名字,如果没有显示提及，默认为 default")
    index_mode:str = pydantic.Field(...,description="用户提及的索引模式，当用户说简单对应的值为 simple，当用户说复杂模式时，对应的值为complex")

class LoadDataAgent(ConversableAgent):      

    DEFAULT_SYSTEM_MESSAGE = "You're a retrieve augmented chatbot."
    
    def __init__(
        self,
        name: str,
        llm: ByzerLLM,        
        retrieval: ByzerRetrieval,        
        chat_name:str,
        owner:str,                
        update_context_retry: int = 3,        
        system_message: Optional[str] = DEFAULT_SYSTEM_MESSAGE,        
        is_termination_msg: Optional[Callable[[Dict], bool]] = None,
        max_consecutive_auto_reply: Optional[int] = None,
        human_input_mode: Optional[str] = "NEVER",
        code_execution_config: Optional[Union[Dict, bool]] = False,
        **kwargs,
    ):       
        super().__init__(
            name,
            llm,retrieval,
            system_message,
            is_termination_msg,
            max_consecutive_auto_reply,
            human_input_mode,
            code_execution_config=code_execution_config,            
            **kwargs,
        )
        self.chat_name = chat_name
        self.owner = owner
        

        self._reply_func_list = []        
        self.register_reply([Agent, ClientActorHandle,str], LoadDataAgent.generate_load_reply) 
        self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.check_termination_and_human_reply)         

        self.llm = llm
        self.retrieval = retrieval
                                    

    def generate_load_reply(
        self,
        raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Union[ClientActorHandle,Agent,str]] = None,
        config: Optional[Any] = None,
    ) -> Tuple[bool, Union[str, Dict, None,ChatResponse]]:  
        
        if messages is None:
            messages = self._messages[get_agent_name(sender)]
                
        new_message = messages[-1]
        content = new_message["content"]

        @self.llm.response()
        def extract_data(s:str)->AgentData:
            pass

        agent_data = extract_data(content)
        id = str(uuid.uuid4())        

        if not agent_data.path:
            def gen():            
                yield ("Please provide a valid path.",None)                               
            return self.stream_reply(gen(),contexts=[])           

        if not agent_data.namespace:
            agent_data.namespace = "default"     

        print(agent_data,flush=True)           

        service_context = get_service_context(self.llm)
        storage_context = get_storage_context(self.llm,self.retrieval,chunk_collection=agent_data.namespace,namespace=agent_data.namespace)

        retrieval_client = SimpleRetrieval(llm=self.llm,retrieval=self.retrieval)
        retrieval_client.delete_from_doc_collection(agent_data.namespace)
        retrieval_client.delete_from_chunk_collection(agent_data.namespace)

        documents = SimpleDirectoryReader(agent_data.path).load_data()

        sp = SentenceSplitter(chunk_size=1024, chunk_overlap=0)        

        nodes = sp.get_nodes_from_documents(
            documents, show_progress=True
        )
        _ = VectorStoreIndex(nodes=nodes, storage_context=storage_context, service_context=service_context)
        
        def gen():                    
            yield (f"Data loaded successfully with mode {agent_data.index_mode}.",None)        
        self.put_stream_reply(id,gen())
        return True, {
            "content":id,
            "metadata":{"agent":self.name,"TERMINATE":True,"stream":True,"stream_id":id,"contexts":[]}
        }                 
        
                
        
                    


from typing import List
import numpy as np
from .generation import GenerateModel, SAMPLE_RATE
from .api import VoiceGenerateAPI

ZH_SPEAKER = "v2/zh_speaker_3"
EN_SPEAKER = "v2/en_speaker_6"

silence = np.zeros(int(0.25 * SAMPLE_RATE))


def is_chinese(t: str) -> bool:
    _is_chinese = True
    l = t.split(" ")
    element_num_15 = len([s for s in l if len(s.strip()) < 15])
    if element_num_15 / len(l) > 0.9:
        _is_chinese = False
    return _is_chinese


def raw_sentence(t: str) -> List[str]:
    import re

    pattern = (
        r"[。；：?！.:;?!]"  # regular expression pattern to match the punctuation marks
    )
    # split the string based on the punctuation marks
    segments = [s for s in re.split(pattern, t) if len(s.strip()) > 0]
    return segments


def not_toolong(t: str) -> List[str]:
    # split the string by white space
    # check the length of every segment which is greater than 20, get the count
    # use the count to divide the length of the list
    l = t.split(" ")
    _is_chinese = is_chinese(t)

    if _is_chinese and len(t) > 30:
        import re

        pattern = r"[。；：，?！,.:;?!]"
        segments = [s for s in re.split(pattern, t) if len(s.strip()) > 0]
        return segments
    elif not _is_chinese and len(l) > 30:
        import re

        pattern = r"[。；：，?！,.:;?!]"
        segments = [s for s in re.split(pattern, t) if len(s.strip()) > 0]
        return segments
    else:
        return [t]


class Inference:
    def __init__(self, model: GenerateModel) -> None:
        self.model = model
        self.api = VoiceGenerateAPI(model)
        self.model.preload_models()
        self.texts = []

    def add(self, text: str, prompt: str) -> None:
        self.texts.append((text, prompt))

    def generate(self, text: str, prompt: str) -> np.ndarray:
        audio_array = self.api.generate_audio(text, history_prompt=prompt)
        return np.concatenate([audio_array, silence.copy()])

    def batch_generate(self) -> np.ndarray:
        # later, we will use Ray to parallelize the generation
        pieces = []
        for text, prompt in self.texts:
            audio_array = self.api.generate_audio(text, history_prompt=prompt)
            pieces += [audio_array, silence.copy()]
        return np.concatenate(pieces)

    def text_to_voice(self, t1: str) -> np.ndarray:
        from langdetect import detect as detect_language

        t = t1.replace("\n", "")
        segments = []
        for s in raw_sentence(t):
            ss = not_toolong(s)
            for sss in ss:
                if len(sss.strip()) > 0:
                    segments.append(sss)
        temps = []
        for s in segments:
            lang = ""
            try:
                lang = detect_language(s)
            except Exception:
                pass

            speaker = ZH_SPEAKER
            if lang == "en":
                speaker = EN_SPEAKER

            print(f"{speaker} will speek: {s}")
            temps.append((s, speaker))

        results = []
        for temp in temps:
            print(f"temp: {temp}")
            import time

            # 统计耗时
            # 记录开始时间
            start_time = time.time()
            result = self.generate(*temp)
            end_time = time.time()
            # 计算执行时间
            execution_time = end_time - start_time
            # 打印执行时间
            print("代码执行时间：", execution_time, "秒")
            results.append(result)

        return np.concatenate(results)


def build_void_infer(model_dir: str, tokenizer_dir: str) -> Inference:
    model = GenerateModel(model_dir=model_dir, tokenizer_dir=tokenizer_dir)
    infer = Inference(model)
    return infer



from ..conversable_agent import ConversableAgent
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from ....utils.client import ByzerLLM,message_utils,code_utils
from byzerllm.utils.retrieval import ByzerRetrieval
from ..agent import Agent
from ray.util.client.common import ClientActorHandle, ClientObjectRef
from .. import get_agent_name,run_agent_func,ChatResponse
import json
try:
    from termcolor import colored
except ImportError:

    def colored(x, *args, **kwargs):
        return x
    
class SQLReviewerAgent(ConversableAgent): 
    DEFAULT_SYSTEM_MESSAGE='''你的任务是检查用户发送给你的Spark SQL语句。
请仔细阅读用户发给你 Spark SQL，具体请按如下方式步骤进行检查，务必一步一步来：

1. 仔细检查SQL中的每一个字段和别名是否都使用反引号（backtick）或者反撇号（grave accent）`括起来。
2. SQL中不允许有文字描述提到让用户用户手动输入的内容，比如 "请输入"，"请填写" 等。
3. SQL中不允许出现诸如 ?, ?, ..., ? 这种参数化查询
4. SQL中不需有诸如 "有取 M 条记录， Top N类" 这种描述性的内容

请输出每一步检查结果，并说明原因。  
                                          
最后，对上面的检查结果和原因重新以json数组格式输出：

```json
[{
    id： 检查步骤序号,
    pass: 是否通过，true/false,
    reason: 原因                                    
}]                      
```
'''
    def __init__(
        self,
        name: str,
        llm: ByzerLLM,        
        retrieval: ByzerRetrieval, 
        chat_name:str,
        owner:str,               
        system_message: Optional[str] = DEFAULT_SYSTEM_MESSAGE,        
        is_termination_msg: Optional[Callable[[Dict], bool]] = None,
        max_consecutive_auto_reply: Optional[int] = None,
        human_input_mode: Optional[str] = "NEVER",
        code_execution_config: Optional[Union[Dict, bool]] = False,
        **kwargs,
    ):       
        super().__init__(
            name,
            llm,retrieval,
            system_message,
            is_termination_msg,
            max_consecutive_auto_reply,
            human_input_mode,
            code_execution_config=code_execution_config,            
            **kwargs,
        )

        self._reply_func_list = []
        # self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.generate_llm_reply)   
        self.register_reply([Agent, ClientActorHandle,str], SQLReviewerAgent.generate_review_reply) 
        self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.check_termination_and_human_reply) 

    def generate_review_reply(
        self,
        raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Union[ClientActorHandle,Agent,str]] = None,
        config: Optional[Any] = None,
        ) -> Tuple[bool, Union[str, Dict, None,ChatResponse]]:  

        if messages is None:
            messages = self._messages[get_agent_name(sender)]

        message = messages[-1]                
        v = self.llm.chat_oai(conversations=message_utils.padding_messages_merge(self._system_message + messages[-1:]))
        print(f'''
review code:
    {message}. 
review result:
    {v[0].output}
''',flush=True)
        checks = json.loads(code_utils.extract_code(v[0].output)[-1][1])

        c = []
        for check in checks:
            if not check["pass"]:
                c.append(check["reason"])

        if len(c) > 0:
            t = "\n".join(c)
            new_message = {"content":f'''代码存在一些问题，具体的问题如下：\n{t}''',"metadata":{}}
            return True, message_utils.copy_error_count(message,new_message)
        
        t = self.llm.chat_oai(conversations=[
    {"role":"user",
    "content":f'''仔细检查下面的 Spark SQL：

{message["content"]}

请找出所有的字段以及别名，去掉函数，保留反引号，并将他们按照出现顺序，以json数组格式输出：

```json
[
  "字段或者别名"
]
```
'''}])
        try:
            fields = json.loads(code_utils.extract_code(t[0].output)[-1][1])
            for field in fields:
                for field in fields:            
                    if "`" not in field:
                        if f"`{field}`" not in message["content"]:
                            new_message = {"content":f'''代码存在问题，字段或者别名: {field} 没有被反引号括起来,请修改''',"metadata":{}}            
                            return True, message_utils.copy_error_count(message,new_message)
        except Exception:
            pass
        
        new_message = {"content":f'''代码没有问题，可以正常允许。''',"metadata":{}}                        
                     
        return True, message_utils.copy_error_count(message,new_message)

        


import os
from os.path import expanduser
import urllib.request
import tarfile
from loguru import logger
import json
from rich.console import Console
from rich.panel import Panel
from rich.progress import Progress
from rich import print as rprint

from byzerllm.apps.byzer_storage.env import get_latest_byzer_retrieval_lib
from byzerllm.apps.byzer_storage import env

console = Console()

class StorageSubCommand:
        
    @staticmethod
    def install(args):
        version = args.version
        home = expanduser("~")        
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")
        libs_dir = os.path.join(base_dir, "storage", "libs", f"byzer-retrieval-lib-{version}")
        if os.path.exists(libs_dir):
            print(f"Byzer Storage version {version} already installed.")
            return
        
        if not os.path.exists(base_dir):
            os.makedirs(base_dir,exist_ok=True)            

        env_info = env.detect_env()

        logger.info("Current environment:")
        logger.info(env_info)
        
        if env_info.java_version == "" or int(env_info.java_version) < 21:
            logger.info("JDK 21 not found, downloading and installing JDK 21...")
            try:
                env.download_and_install_jdk21(env_info, base_dir)
            except Exception as e:
                logger.error(f"Error downloading and installing JDK 21: {str(e)}. You may need to install JDK 21 manually.")
                        
        download_url = f"https://download.byzer.org/byzer-retrieval/byzer-retrieval-lib-{version}.tar.gz"
        libs_dir = os.path.join(base_dir, "storage", "libs")
        
        os.makedirs(libs_dir, exist_ok=True)
        download_path = os.path.join(libs_dir, f"byzer-retrieval-lib-{version}.tar.gz")
        if os.path.exists(download_path):
            logger.info(f"Byzer Storage version {version} already downloaded.")
        else:             
            def download_with_progressbar(url, filename):
                def progress(count, block_size, total_size):
                    percent = int(count * block_size * 100 / total_size)
                    print(f"\rDownload progress: {percent}%", end="")
            
                urllib.request.urlretrieve(url, filename,reporthook=progress)                
  
            logger.info(f"Download Byzer Storage version {version}: {download_url}")
            download_with_progressbar(download_url, download_path)    

            with tarfile.open(download_path, "r:gz") as tar:
                tar.extractall(path=libs_dir)
        
        print("Byzer Storage installed successfully")

    def collection(args):        
        from byzerllm.apps.llama_index.collection_manager import CollectionManager, CollectionItem
        home = expanduser("~")        
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")
        collection_manager = CollectionManager(base_dir)
        if args.name:            
            collection = CollectionItem(name=args.name, description=args.description)
            collection_manager.add_collection(collection)
            print(f"Collection {args.name} added successfully.")
        else:
            print("Please provide collection name.")    
    
    @staticmethod
    def start(args):
        import byzerllm
        from byzerllm.utils.retrieval import ByzerRetrieval
        version = args.version
        cluster = args.cluster
        home = expanduser("~")        
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")
        
        with console.status("[bold green]Starting Byzer Storage...") as status:
            libs_dir = os.path.join(base_dir, "storage", "libs", f"byzer-retrieval-lib-{version}")
            data_dir = os.path.join(base_dir, "storage", "data")

            if not os.path.exists(os.path.join(data_dir,cluster)):
                os.makedirs(data_dir,exist_ok=True)
                rprint("[green]✓[/green] Created data directory")

            if not os.path.exists(libs_dir):            
                StorageSubCommand.install(args)
                rprint("[green]✓[/green] Installed Byzer Storage")

            code_search_path = [libs_dir]
            
            status.update("[bold blue]Connecting to cluster...")
            env_vars = byzerllm.connect_cluster(address=args.ray_address,code_search_path=code_search_path)
            rprint("[green]✓[/green] Connected to cluster")
            
            retrieval = ByzerRetrieval()
            retrieval.launch_gateway()
            rprint("[green]✓[/green] Launched gateway")

            if retrieval.is_cluster_exists(name=cluster):
                console.print(Panel(f"[yellow]Cluster {cluster} already exists. Please stop it first.[/yellow]"))
                return 
            
            base_model_dir = os.path.join(base_dir, "storage","models")
            os.makedirs(base_model_dir,exist_ok=True)
            bge_model = os.path.join(base_model_dir,"AI-ModelScope","bge-large-zh")
            
            status.update("[bold blue]Downloading embedding model...")
            from modelscope.hub.snapshot_download import snapshot_download
            import huggingface_hub
            if not os.path.exists(bge_model):
                model_path = snapshot_download(
                    model_id="AI-ModelScope/bge-large-zh",
                    cache_dir=base_model_dir,
                    local_files_only=huggingface_hub.constants.HF_HUB_OFFLINE                
                )
            else:
                model_path = bge_model        
            rprint(f"[green]✓[/green] Embedding model downloaded: {model_path}")

            llm = byzerllm.ByzerLLM()        
            if not llm.is_model_exist("emb"): 
                status.update("[bold blue]Deploying embedding model...")
                from byzerllm.utils.client import InferBackend         
                llm.setup_num_workers(1).setup_infer_backend(InferBackend.Transformers)
                llm.setup_gpus_per_worker(0).setup_cpus_per_worker(0.01).setup_worker_concurrency(20)
                llm.deploy(
                    model_path=bge_model,
                    pretrained_model_type="custom/bge",
                    udf_name="emb",            
                    infer_params={}
                )               
                rprint("[green]✓[/green] Deployed embedding model")
            
            cluster_json = os.path.join(base_dir, "storage", "data",f"{cluster}.json")
            if os.path.exists(cluster_json):
                StorageSubCommand.restore(args)
                console.print(Panel("[green]Byzer Storage restored and started successfully[/green]"))
                return 
                
            status.update("[bold blue]Starting cluster...")
            builder = retrieval.cluster_builder()
            builder.set_name(cluster).set_location(data_dir).set_num_nodes(1).set_node_cpu(1).set_node_memory("2g")
            builder.set_java_home(env_vars["JAVA_HOME"]).set_path(env_vars["PATH"]).set_enable_zgc()
            builder.start_cluster()
            
            with open(os.path.join(base_dir, "storage", "data",f"{cluster}.json"),"w") as f:
                f.write(json.dumps(retrieval.cluster_info(cluster),ensure_ascii=False))
            
            console.print(Panel("[green]Byzer Storage started successfully[/green]"))
        

    @staticmethod 
    def stop(args):    
        import byzerllm
        from byzerllm.utils.retrieval import ByzerRetrieval
        version = args.version
        cluster = args.cluster
        home = expanduser("~")        
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")
        libs_dir = os.path.join(base_dir, "storage", "libs", f"byzer-retrieval-lib-{version}")
        cluster_json = os.path.join(base_dir, "storage", "data",f"{cluster}.json")        

        if not os.path.exists(cluster_json) or not os.path.exists(libs_dir):
            print("No instance find.")
            return

        code_search_path = [libs_dir]
        
        logger.info(f"Connect and start Byzer Retrieval version {version}")
        byzerllm.connect_cluster(address=args.ray_address,code_search_path=code_search_path)             
        retrieval = ByzerRetrieval()
        retrieval.launch_gateway()
        retrieval.shutdown_cluster(cluster_name=cluster)

    @staticmethod 
    def export(args):   
        import byzerllm
        from byzerllm.utils.retrieval import ByzerRetrieval
        version = args.version
        cluster = args.cluster
        home = expanduser("~")        
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")
        libs_dir = os.path.join(base_dir, "storage", "libs", f"byzer-retrieval-lib-{version}")
        cluster_json = os.path.join(base_dir, "storage", "data",f"{cluster}.json")        

        if not os.path.exists(cluster_json) or not os.path.exists(libs_dir):
            print("No instance find.")
            return

        code_search_path = [libs_dir]
        
        logger.info(f"Connect and restore Byzer Retrieval version {version}")
        byzerllm.connect_cluster(address=args.ray_address,code_search_path=code_search_path)        
     
        retrieval = ByzerRetrieval()
        retrieval.launch_gateway()
        
        with open(cluster_json,"w") as f:
            f.write(json.dumps(retrieval.cluster_info(cluster),ensure_ascii=False))

        print(f"Byzer Storage export successfully. Please check {cluster_json}")    


    
    def restore(args):
        import byzerllm
        from byzerllm.utils.retrieval import ByzerRetrieval
        version = args.version
        cluster = args.cluster
        home = expanduser("~")        
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")
        libs_dir = os.path.join(base_dir, "storage", "libs", f"byzer-retrieval-lib-{version}")
        cluster_json = os.path.join(base_dir, "storage", "data",f"{cluster}.json")

        if not os.path.exists(cluster_json) or not os.path.exists(libs_dir):
            print("No instance find.")
            return

        code_search_path = [libs_dir]
        
        logger.info(f"Connect and restore Byzer Retrieval version {version}")
        byzerllm.connect_cluster(address=args.ray_address,code_search_path=code_search_path)        
     
        retrieval = ByzerRetrieval()
        retrieval.launch_gateway()

        if not retrieval.is_cluster_exists(cluster):
            with open(cluster_json,"r") as f:
                cluster_info = f.read()
            
            retrieval.restore_from_cluster_info(json.loads(cluster_info))
            
            print("Byzer Storage restore successfully")
        else:
            print(f"Cluster {cluster} is already exists")




from ..conversable_agent import ConversableAgent
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from ....utils.client import ByzerLLM,message_utils
from byzerllm.utils.retrieval import ByzerRetrieval
from ..agent import Agent
from ray.util.client.common import ClientActorHandle, ClientObjectRef
import time
from .. import get_agent_name,run_agent_func,ChatResponse,modify_last_message,modify_message_content
from langchain import PromptTemplate
import json


class PreviewFileAgent(ConversableAgent):    

    DEFAULT_SYSTEM_MESSAGE = """You are a helpful AI assistant. You will use your knowledge to help the user to preview the file.Try
to use Python and Pandas to read the file and show the first 5 rows of the file. The user will mention the file path in his/her question.
The packages all are installed, you can use it directly.
Try to generate python code which should match the following requirements:
1. try to read the file according the suffix of file name in Try block
2. if read success, set variable loaded_successfully to True, otherwise set it to False.
3. if loaded_successfully is True, then assigh the loaded data with head() to file_preview, otherwise assign error message to file_preview
4. make sure the loaded_successfully, file_preview are defined in the global scope
"""    
    
    DEFAULT_USER_MESSAGE = """
We have a file, the path is: {file_path}. Try to  write code preview this file. Make sure the {file_path} is defined in the code. We need to 
execute the code to preview the file. If the code is correct, the file will be loaded successfully and the first 5 rows of the file will be shown.
"""

    def __init__(
        self,
        name: str,
        llm: ByzerLLM,
        retrieval: ByzerRetrieval,
        chat_name:str,
        owner:str,
        code_agent: Union[Agent, ClientActorHandle,str],
        system_message: Optional[str] = DEFAULT_SYSTEM_MESSAGE,        
        is_termination_msg: Optional[Callable[[Dict], bool]] = None,
        max_consecutive_auto_reply: Optional[int] = None,
        human_input_mode: Optional[str] = "NEVER",
        code_execution_config: Optional[Union[Dict, bool]] = False,
        **kwargs,
    ):
        super().__init__(
            name,
            llm,retrieval,
            system_message,
            is_termination_msg,
            max_consecutive_auto_reply,
            human_input_mode,
            code_execution_config=code_execution_config,            
            **kwargs,
        )
        self.code_agent = code_agent
        self._reply_func_list = []
        # self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.generate_llm_reply)   
        self.register_reply([Agent, ClientActorHandle,str], PreviewFileAgent.generate_code_reply) 
        self.register_reply([Agent, ClientActorHandle,str], PreviewFileAgent.reply_python_code_agent) 
        self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.check_termination_and_human_reply) 

    def reply_python_code_agent(
        self,
        raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Union[ClientActorHandle,Agent,str]] = None,
        config: Optional[Any] = None,
    ) -> Tuple[bool, Union[str, Dict, None,ChatResponse]]: 
        if get_agent_name(sender) != get_agent_name(self.code_agent):
            return False, None
        
        if messages is None:
            messages = self._messages[get_agent_name(sender)]

        message = messages[-1]
        raw_message = messages[-1]["metadata"]["execute_result"]

        if raw_message.status == 0 and "loaded_successfully" in raw_message.variables and raw_message.variables["loaded_successfully"]:                                 
            # stop the conversation if the code agent gives the success message
            return True, None
        else:
            print(raw_message,flush=True)
            if message_utils.check_error_count(message,3):
                return True, {
                    "content":f'''Fail to load the file: {raw_message.variables.get("file_path","")}. reason: {raw_message.output}''' + "\nTERMINATE",
                    "metadata":{"TERMINATE":True,"code":1}
                }
            # the code may be wrong, so generate a new code according to the conversation so far 
            extra_messages = []            
            if "loaded_successfully" not in raw_message.variables:                
                extra_messages.append(self.create_temp_message("loaded_successfully is not defined"))
            
            elif raw_message.variables["loaded_successfully"] is False:                            
                extra_messages.append(self.create_temp_message("loaded_successfully is False, it means the file is not loaded successfully, check the file path and the code then try again"))
                

            _,code = self.generate_llm_reply(raw_message,messages + extra_messages,sender)
            m = self.create_temp_message(code)
            message_utils.copy_error_count(message,m)

            return True, message_utils.inc_error_count(m)
        
        
    
    def generate_code_reply(
        self,
        raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Union[ClientActorHandle,Agent,str]] = None,
        config: Optional[Any] = None,
    ) -> Tuple[bool, Union[str, Dict, None,ChatResponse]]:  
        
        if messages is None:
            messages = self._messages[get_agent_name(sender)]
        
        new_message = messages[-1] 
        file_path = new_message["metadata"]["file_path"]
        content = PromptTemplate.from_template(self.DEFAULT_USER_MESSAGE).format(file_path=new_message["metadata"]["file_path"])
        new_messages = modify_last_message(messages,modify_message_content(new_message,content))
        
        _,code = self.generate_llm_reply(raw_message,new_messages,sender)            
        
        # only the first time we should keep the message sent to the code agent which have file_path, file_ref
        # in message metadata, when the sandbox is created, then we will reuse the sandbox, no need to contain
        # the file_path, file_ref in the message metadata.         
        self.send(message=self.create_temp_message(code,new_message),recipient=self.code_agent)

        # get the code agent's reply
        last_message = self._messages[get_agent_name(self.code_agent)][-1]
        if last_message["metadata"].get("code",0) != 0:
            return True, {"content":f'''Fail to load the file: {file_path}. reason: {response.variables.get("file_preview","")}''',"metadata":{"TERMINATE":True,"code":1}}
        
        response:ChatResponse = last_message["metadata"]["execute_result"]

        if "loaded_successfully" not in response.variables or response.variables["loaded_successfully"] is False:
            return True, {"content":f'''Fail to load the file: {file_path}. reason: {response.variables.get("file_preview","")}''',"metadata":{"TERMINATE":True,"code":1}}
        
        file_preview = response.variables["file_preview"].to_csv(index=False)    
        
        return True, {"content":file_preview,"metadata":{"TERMINATE":True}}
            
            
        

        
        
    def create_temp_message(self,code,original_message=None):
        temp_message = {
            "content":code,
            "metadata":{
                "target_names":{"loaded_successfully":None,"file_preview":None}
            },
            "role":"user"
        }
        if original_message is not None:
            temp_message["metadata"] = {**original_message["metadata"],**temp_message["metadata"]}
        return temp_message    
        


import json

from langchain import PromptTemplate
import ray
import os
from typing import Dict, List, Any, Tuple
import uuid
from ray.util.client.common import ClientActorHandle, ClientObjectRef
from pyjava.storage  import streaming_tar
import time

from . import BuilderParams,QueryParams
from .qa_strategy import FullDocCombineFormatFactory, DocRetrieveStrategyFactory
from .vector_db import VectorDB
from .client import ByzerLLMClient


@ray.remote
class ByzerLLMQAQueryWorker:
    def __init__(self,refs:List[ClientObjectRef],client:ByzerLLMClient,query_param:QueryParams) -> None:        
        self.client = client        
        db_path = os.path.join(query_param.local_path_prefix,str(uuid.uuid4()))
        streaming_tar.save_rows_as_file((ray.get(ref) for ref in refs),db_path)
        self.db = VectorDB(db_path,self.client)
                
    def query(self,prompt:str,q:str,k=4):                                
        return self.db.query(prompt,q,k)        



class ByzerLLMQA:
    def __init__(self,db_dir:str,client:ByzerLLMClient,query_params:QueryParams) -> None:
        self.db_dir = db_dir
        self.client = client
        self.query_params = query_params
        self.db_dirs = [os.path.join(self.db_dir,item) for item in os.listdir(self.db_dir)]
        self.dbs = []
        for dd in self.db_dirs:
            refs = [ray.put(item) for item in streaming_tar.build_rows_from_file(dd)]
            self.dbs.append(ByzerLLMQAQueryWorker.remote(refs,self.client,self.query_params))                    

    def query(self,prompt:str,q:str,k=4,hint="",input:Dict[str,Any]={}): 
         
        docs_with_score = [] 

        time1 = time.time()
        submit_q = [db.query.remote("",q,k) for db in self.dbs]        
        for q_func in submit_q:            
            t = ray.get(q_func)
            docs_with_score = docs_with_score + t

        query_vector_db_time = time.time() - time1        
        print(f"VectorDB query time taken:{query_vector_db_time}s. total chunks: {len(docs_with_score)}")   

        docs = sorted(docs_with_score, key=lambda doc: doc[1],reverse=False)

        strategy = input.get("strategy", "")
        docs = DocRetrieveStrategyFactory(strategy).retrieve(docs, k)

        if hint == "show_only_context":
            return json.dumps([{"score":float(doc[1]),"content":doc[0].page_content} for doc in docs[0:k]],ensure_ascii=False,indent=4)
         
        newq , temp_metas = FullDocCombineFormatFactory(input).combine(docs, k)
        
        show_full_query  = hint == "show_full_query"         

        if not prompt:
            prompt = "{context} \n {query}"

        prompt_template = PromptTemplate.from_template(prompt)
 
        final_query = prompt_template.format(context=newq, query=q)

        time2 = time.time()
        
        top_p=float(input.get("top_p",0.7))
        temperature=float(input.get("temperature",0.9))
        max_length = int(input.get("max_length", 1024))

        v = self.client.chat(final_query,[],extra_query={"top_p":top_p,"temperature":temperature, "max_length":max_length})
        chat_time = time.time() - time2

        if show_full_query:
          return json.dumps({
             "query": final_query,
             "predict": v,
             "vectordb_time": f"{query_vector_db_time}s",
             "chat_time": f"{chat_time}s",
             "metas":temp_metas
          },ensure_ascii=False,indent=4)
        else:
          return v

    def predict(self,input:Dict[str,Any]):        
        q = input["instruction"]
        prompt = input.get("prompt","")
        hint = input.get("hint","")
        k = int(input.get("k","4"))
        return self.query(prompt,q,k,hint,input)

@ray.remote
class RayByzerLLMQAWorker: 
    def __init__(self,data_ref,client:ByzerLLMClient) -> None:
        self.data_ref = data_ref        
        self.client = client         
    
    def build(self,params:BuilderParams,extra_params={}):
        from pyjava.api.mlsql import RayContext
        from pyjava.storage import streaming_tar
        import uuid
        import shutil        

        data_path = os.path.join(params.local_path_prefix,str(uuid.uuid4()))
        
        if not os.path.exists(data_path):
            os.makedirs(data_path, exist_ok=True)
            
        with open(os.path.join(data_path,"data.json"),"w",encoding="utf-8") as f:
            for item in RayContext.collect_from([self.data_ref]):
                f.write(json.dumps(item,ensure_ascii=False)+"\n")
        
        
        db_dir = os.path.join(params.local_path_prefix,str(uuid.uuid4()))
        db = VectorDB(db_dir,self.client,extra_params=extra_params)
        db.build_from(data_path,params,extra_params=extra_params)
        
        refs = []
        for item in  streaming_tar.build_rows_from_file(db_dir):
            item_ref = ray.put(item)
            refs.append(item_ref)

        shutil.rmtree(data_path,ignore_errors=True)
        shutil.rmtree(db_dir)
        return refs

# QA Vector Store Builder
class RayByzerLLMQA:
    def __init__(self,client:ByzerLLMClient) -> None:
        self.db_dir = ""
        self.client = client        
    
    def save(self,data_refs,params=BuilderParams(),builder_params={}):        
        from pyjava.storage import streaming_tar     

        self.db_dir = os.path.join(params.local_path_prefix,"qa",str(uuid.uuid4()))

        data = []
        workers = []

        ## build vector db file in parallel
        print(f"Start {len(data_refs)} workers to build vector db")
        for data_ref in data_refs:            
            worker = RayByzerLLMQAWorker.remote(data_ref,self.client)
            workers.append(worker)
            build_func = worker.build.remote(params,builder_params)
            data.append(build_func)

        ## gather all db file and merge into one
        print(f"gather all db file and merge into one {self.db_dir}")
        for index, build_func in enumerate(data):
            sub_data = ray.get(build_func)
            temp_dir = os.path.join(self.db_dir,f"vecdb_{index}")
            streaming_tar.save_rows_as_file((ray.get(ref) for ref in sub_data),temp_dir)                     
        
        return streaming_tar.build_rows_from_file(self.db_dir)
        









import transformers
from transformers import (
    PreTrainedModel,
    TrainingArguments,
    DataCollator,
    PreTrainedTokenizerBase,
    EvalPrediction,
    TrainerCallback,
)
from typing import Callable, Dict, List, Optional, Tuple, Union, Any
from torch import nn
from torch.utils.data import Dataset, DataLoader
from transformers.utils import (
    logging,
)
from typing import Optional
import os
import torch


logger = logging.get_logger(__name__)

# Name of the files used for checkpointing
TRAINING_ARGS_NAME = "training_args.bin"
TRAINER_STATE_NAME = "trainer_state.json"
OPTIMIZER_NAME = "optimizer.pt"
SCHEDULER_NAME = "scheduler.pt"
SCALER_NAME = "scaler.pt"


class Trainer(transformers.Trainer):
    """
    主要修改逻辑：通过传入compute_loss，支持自定义loss计算方式
    """
    def __init__(
            self,
            model: Union[PreTrainedModel, nn.Module] = None,
            args: TrainingArguments = None,
            data_collator: Optional[DataCollator] = None,
            train_dataset: Optional[Dataset] = None,
            eval_dataset: Optional[Dataset] = None,
            tokenizer: Optional[PreTrainedTokenizerBase] = None,
            model_init: Callable[[], PreTrainedModel] = None,
            compute_metrics: Optional[Callable[[EvalPrediction], Dict]] = None,
            callbacks: Optional[List[TrainerCallback]] = None,
            optimizers: Tuple[torch.optim.Optimizer, torch.optim.lr_scheduler.LambdaLR] = (None, None),
            preprocess_logits_for_metrics: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = None,
            compute_loss=None,
    ):
        super(Trainer, self).__init__(
            model=model,
            args=args,
            data_collator=data_collator,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            tokenizer=tokenizer,
            model_init=model_init,
            compute_metrics=compute_metrics,
            callbacks=callbacks,
            optimizers=optimizers,
            preprocess_logits_for_metrics=preprocess_logits_for_metrics,
        )
        self.loss_func = compute_loss

    def compute_loss(self, model, inputs, return_outputs=False):
        """
        重写loss的计算方式
        How the loss is computed by Trainer. By default, all models return the loss in the first element.

        Subclass and override for custom behavior.
        """
        return self.loss_func(model, inputs, self.args, return_outputs)


class LoRATrainer(Trainer):
    """
    修改checkkpoint的保存逻辑，只保存lora
    """
    def _save(self, output_dir: Optional[str] = None, state_dict=None):
        # If we are executing this function, we are the process zero, so we don't check for that.
        output_dir = output_dir if output_dir is not None else self.args.output_dir
        os.makedirs(output_dir, exist_ok=True)
        logger.info(f"Saving model checkpoint to {output_dir}")
        # 保存lora权重和配置
        self.model.save_pretrained(
            output_dir, state_dict=state_dict, safe_serialization=self.args.save_safetensors
        )

        if self.tokenizer is not None:
            self.tokenizer.save_pretrained(output_dir)

        # Good practice: save your training arguments together with the trained model
        torch.save(self.args, os.path.join(output_dir, TRAINING_ARGS_NAME))


from .conversable_agent import ConversableAgent
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from ...utils.client import ByzerLLM,message_utils
from byzerllm.utils.retrieval import ByzerRetrieval
from .agent import Agent
from ray.util.client.common import ClientActorHandle, ClientObjectRef
import time
from . import get_agent_name,run_agent_func,ChatResponse


class AssistantAgent(ConversableAgent):    

    DEFAULT_SYSTEM_MESSAGE = """You are a helpful AI assistant.
Solve tasks using your coding and language skills.
In the following cases, suggest python code (in a python coding block) or shell script (in a sh coding block) for the user to execute.
    1. When you need to collect info, use the code to output the info you need, for example, browse or search the web, download/read a file, print the content of a webpage or a file, get the current date/time, check the operating system. After sufficient info is printed and the task is ready to be solved based on your language skill, you can solve the task by yourself.
    2. When you need to perform some task with code, use the code to perform the task and output the result. Finish the task smartly.
Solve the task step by step if you need to. If a plan is not provided, explain your plan first. Be clear which step uses code, and which step uses your language skill.
When using code, you must indicate the script type in the code block. The user cannot provide any other feedback or perform any other action beyond executing the code you suggest. The user can't modify your code. So do not suggest incomplete code which requires users to modify. Don't use a code block if it's not intended to be executed by the user.
If you want the user to save the code in a file before executing it, put # filename: <filename> inside the code block as the first line. Don't include multiple code blocks in one response. Do not ask users to copy and paste the result. Instead, use 'print' function for the output when relevant. Check the execution result returned by the user.
If the result indicates there is an error, fix the error and output the code again. Suggest the full code instead of partial code or code changes. If the error can't be fixed or if the task is not solved even after the code is executed successfully, analyze the problem, revisit your assumption, collect additional info you need, and think of a different approach to try.
When you find an answer, verify the answer carefully. Include verifiable evidence in your response if possible.

注意：
1. 不要提供类似 if __name__ == '__main__' 的判断。
2. 代码需要通过 exec 函数执行，所以不要使用 return 语句。
3. 不要使用 input 这种需要用户输入的函数。
"""

    def __init__(
        self,
        name: str,
        llm: ByzerLLM,
        retrieval: ByzerRetrieval,
        chat_name:str,
        owner:str,
        code_agent: Union[Agent, ClientActorHandle,str],
        system_message: Optional[str] = DEFAULT_SYSTEM_MESSAGE,        
        is_termination_msg: Optional[Callable[[Dict], bool]] = None,
        max_consecutive_auto_reply: Optional[int] = None,
        human_input_mode: Optional[str] = "NEVER",
        code_execution_config: Optional[Union[Dict, bool]] = False,
        **kwargs,
    ):
        super().__init__(
            name,
            llm,retrieval,
            system_message,
            is_termination_msg,
            max_consecutive_auto_reply,
            human_input_mode,
            code_execution_config=code_execution_config,            
            **kwargs,
        )
        self.code_agent = code_agent
        self._reply_func_list = []
        # self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.generate_llm_reply)   
        self.register_reply([Agent, ClientActorHandle,str], AssistantAgent.generate_code_reply) 
        self.register_reply([Agent, ClientActorHandle,str], AssistantAgent.reply_python_code_agent) 
        self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.check_termination_and_human_reply) 

    def reply_python_code_agent(
        self,
        raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Union[ClientActorHandle,Agent,str]] = None,
        config: Optional[Any] = None,
    ) -> Tuple[bool, Union[str, Dict, None,ChatResponse]]: 
        if get_agent_name(sender) != get_agent_name(self.code_agent):
            return False, None
        
        if messages is None:
            messages = self._messages[get_agent_name(sender)]  
                    
        raw_message = messages[-1]["metadata"]["execute_result"]
        if raw_message.status == 0:
            # stop the conversation if the code agent gives the success message
            return True, None
        else:
            # the code may be wrong, so generate a new code according to the conversation so far
            if message_utils.check_error_count(messages[-1],3):
                return True, {
                    "content":f'''FAIL TO GNERATE CODE : {raw_message.output}''',
                    "metadata":{"TERMINATE":True,"code":1}
                }
            final,output = self.generate_llm_reply(raw_message,messages,sender)
            temp_message = {
                "content":output,
            }
            message_utils.copy_error_count(messages[-1],temp_message)
            message_utils.inc_error_count(temp_message)
            return True, temp_message

    def generate_code_reply(
        self,
        raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Union[ClientActorHandle,Agent,str]] = None,
        config: Optional[Any] = None,
    ) -> Tuple[bool, Union[str, Dict, None,ChatResponse]]:  
        
        if messages is None:
            messages = self._messages[get_agent_name(sender)]
        
        final,output = self.generate_llm_reply(raw_message,messages,sender)            
                                 
        self.send(message=output,recipient=self.code_agent)

        # summarize the conversation so far  
        last_message = self._messages[get_agent_name(self.code_agent)][-1]                
        if last_message["metadata"].get("code",0) != 0:
            return True, {"content":f'''FAIL TO GNERATE CODE ''',"metadata":{"TERMINATE":True,"code":1}}

        # give the result to the user             
        return True, {"content":last_message["content"],"metadata":{"TERMINATE":True}}
   
            
        

       
        



import io
import os
import uuid
import tempfile

def process_pdf(binnary_data):
    from langchain.document_loaders import PyPDFLoader
    docs = []
    with tempfile.TemporaryDirectory()  as temp_dir:
        pdf_content = io.BytesIO(binnary_data)            
        temp_pdf = os.path.join(temp_dir.name, f"tmp{str(uuid.uuid4())}")

        with open(temp_pdf, 'wb') as f:
            f.write(pdf_content.read())
        
        loader = PyPDFLoader(temp_pdf)
        docs = loader.load()
    
    content = "\n".join([doc.page_content for doc in docs])
    return content


"""Code for moss-moon-003-sft"""

import os
import copy
import json
import torch
import logging
import argparse

from tqdm import tqdm
import torch.distributed as dist
from torch.utils.data import Dataset, DataLoader
from torch.utils.tensorboard import SummaryWriter
from accelerate import Accelerator, DeepSpeedPlugin
from transformers import set_seed, get_cosine_schedule_with_warmup

from models.modeling_moss import MossForCausalLM
from models.tokenization_moss import MossTokenizer


logger = logging.getLogger(__name__)
logging.basicConfig(level='INFO')


class SFTDataset(Dataset):
    def __init__(self, data_dir, tokenizer, data_type='train'):
        super().__init__()

        self.data_dir = data_dir
        self.tokenizer = tokenizer
        self.data_type = data_type

        self.data = []
        # We do not calculate losses for the meta instruction or results returned by plugins
        # The token spans with label -100, [(span_start, span_end), ...]
        self.no_loss_spans = []

        self.load_data()

    def load_data(self):
        logger.info("Loading data...")
        data_file = os.path.join(self.data_dir, f'{self.data_type}_data')
        no_loss_spans_file = os.path.join(self.data_dir, f'{self.data_type}_no_loss_spans')
        if os.path.exists(data_file) and os.path.exists(no_loss_spans_file):
            self.data = torch.load(data_file, map_location='cpu')
            self.no_loss_spans = torch.load(no_loss_spans_file, map_location='cpu')
        else:
            with open(os.path.join(self.data_dir, f'{self.data_type}.jsonl'), 'r') as f:
                for line in f:
                    sample = json.loads(line)

                    chat = sample['chat']
                    num_turns = int(sample['num_turns'])

                    meta_instruction = sample['meta_instruction']
                    instruction_ids = self.tokenizer.encode(meta_instruction)
                    assert isinstance(instruction_ids, list) and len(instruction_ids) > 0
                    
                    input_ids = copy.deepcopy(instruction_ids)
                    no_loss_spans = [(0, len(instruction_ids))]

                    for i in range(num_turns):
                        cur_turn_ids = []
                        cur_no_loss_spans = []
                        cur_turn = chat[f'turn_{i+1}']
                        for key, value in cur_turn.items():

                            cur_ids = self.tokenizer.encode(value)

                            if key == 'Tool Responses':
                                # The format tokens (<|Results|>:...<eor>\n) should have losses. 
                                cur_no_loss_spans.append((len(input_ids + cur_turn_ids) + 5, len(input_ids + cur_turn_ids + cur_ids) - 2))    

                            assert isinstance(cur_ids, list) and len(cur_ids) > 0

                            cur_turn_ids.extend(cur_ids)

                        if len(input_ids + cur_turn_ids) > 2048:
                            break

                        input_ids.extend(cur_turn_ids)
                        no_loss_spans.extend(cur_no_loss_spans)

                    if len(input_ids) == len(instruction_ids):
                        continue

                    assert len(input_ids) > 0 and len(input_ids) <= 2048

                    self.data.append(input_ids)
                    self.no_loss_spans.append(no_loss_spans)
            
            torch.save(self.data, data_file)
            torch.save(self.no_loss_spans, no_loss_spans_file)

        logger.info(f"Load data successfully, total {len(self.data)} training samples")

    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, index):
        data = copy.deepcopy(self.data[index])
        no_loss_spans = copy.deepcopy(self.no_loss_spans[index])
        
        data = torch.tensor(data, dtype=torch.long)
        attn_mask = torch.ones_like(data, dtype=torch.bool)
        label = copy.deepcopy(data)

        for no_loss_span in no_loss_spans:
            label[no_loss_span[0] : no_loss_span[1]] = -100

        return data, attn_mask, label
    
    def collate_fn(self, batch):
        batch_input_ids, batch_attn_mask, batch_labels = [], [], []
        for input_ids, attn_mask, label in batch:
            batch_input_ids.append(input_ids)
            batch_attn_mask.append(attn_mask)
            batch_labels.append(label)

        batch_input_ids = torch.nn.utils.rnn.pad_sequence(batch_input_ids, batch_first=True, padding_value=self.tokenizer.eos_token_id)
        batch_attn_mask = torch.nn.utils.rnn.pad_sequence(batch_input_ids, batch_first=True, padding_value=0).to(torch.bool)
        batch_labels = torch.nn.utils.rnn.pad_sequence(batch_labels, batch_first=True, padding_value=-100)

        return batch_input_ids, batch_attn_mask, batch_labels
    

class SFTMetric:
    def __init__(self, device):
        self.n_step = 0
        self.right = torch.Tensor([0]).to(device=device)
        self.total = torch.Tensor([0]).to(device=device)
        self.total_loss = torch.Tensor([0]).to(device=device)
        self.world_size = dist.get_world_size()

    def __call__(self, logits, labels, loss):
        return self.update(logits, labels, loss)

    def update(self, logits, labels, loss):
        self.n_step += 1
        with torch.no_grad():
            shift_preds = logits[..., :-1, :].argmax(dim=-1)
            shift_labels = labels[..., 1:]
            self.right += (shift_preds == shift_labels).masked_fill(shift_labels.eq(-100), 0).sum().item()
            self.total += (shift_labels != -100).sum().item()
            self.total_loss += loss.item()

    def get_metric(self, reset=True):
        dist.all_reduce(self.right, op=torch.distributed.ReduceOp.SUM)
        dist.all_reduce(self.total, op=torch.distributed.ReduceOp.SUM)
        dist.all_reduce(self.total_loss, op=torch.distributed.ReduceOp.SUM)

        acc = (self.right / self.total).item()
        loss = self.total_loss.item() / (self.world_size * self.n_step)

        if reset:
            self.n_step = 0
            self.right.fill_(0)
            self.total.fill_(0)
            self.total_loss.fill_(0)
        return acc, loss
    

def train(args):

    # deepspeed needs to know your gradient accumulation steps before hand, so don't forget to pass it
    # Remember you still need to do gradient accumulation by yourself, just like you would have done without deepspeed
    # deepspeed_plugin = DeepSpeedPlugin(zero_stage=3, gradient_accumulation_steps=1)
    # deepspeed_plugin.deepspeed_config['train_micro_batch_size_per_gpu'] = 2
    accelerator = Accelerator(mixed_precision='fp16') 

    if accelerator.is_main_process:
        writer = SummaryWriter(args.log_dir)
        writer.add_hparams(vars(args), {})

    accelerator.state.deepspeed_plugin.deepspeed_config['train_micro_batch_size_per_gpu'] = args.train_bsz_per_gpu

    tokenizer = MossTokenizer.from_pretrained(args.model_path)
    model = MossForCausalLM.from_pretrained(args.model_path, use_cache=False)

    model.transformer.gradient_checkpointing = True
    assert model.transformer.gradient_checkpointing is True

    # Optimizer
    # Split weights in two groups, one with weight decay and the other not.
    no_decay = ["bias", "LayerNorm.weight"]
    optimizer_grouped_parameters = [
        {
            "params": [p for n, p in model.named_parameters() if not any(nd in n for nd in no_decay)],
            "weight_decay": args.weight_decay,
        },
        {
            "params": [p for n, p in model.named_parameters() if any(nd in n for nd in no_decay)],
            "weight_decay": 0.0,
        },
    ]

    optimizer = torch.optim.AdamW(optimizer_grouped_parameters, lr=args.learning_rate)

    train_dataset = SFTDataset(args.data_dir, tokenizer)
    train_dataloader = DataLoader(train_dataset, batch_size=args.train_bsz_per_gpu, shuffle=True, drop_last=True, collate_fn=train_dataset.collate_fn)

    val_dataset = SFTDataset(args.data_dir, tokenizer, data_type='val')
    val_dataloader = DataLoader(val_dataset, batch_size=args.eval_bsz_per_gpu, shuffle=False, drop_last=True, collate_fn=train_dataset.collate_fn)

    num_training_steps = (len(train_dataloader) * args.n_epochs) // accelerator.gradient_accumulation_steps
    lr_scheduler = get_cosine_schedule_with_warmup(optimizer, num_warmup_steps=int(args.warmup_rates * num_training_steps), num_training_steps=num_training_steps)

    model, optimizer, train_dataloader, val_dataloader, lr_scheduler = accelerator.prepare(model, optimizer, train_dataloader, val_dataloader, lr_scheduler)

    global_step = 0
    metric = SFTMetric(device=torch.cuda.current_device())

    model.train()
    for epoch in range(args.n_epochs):
        for batch_cnt, (input_ids, attention_mask, labels) in enumerate(train_dataloader):
            if batch_cnt == 1 and epoch == 0:
                torch.cuda.empty_cache()

            optimizer.zero_grad()

            output = model(input_ids=input_ids, attention_mask=attention_mask, labels=labels, return_dict=True)
            loss = output.loss

            metric(output.logits, labels, loss)
            acc, train_loss = metric.get_metric()

            accelerator.backward(loss)
            optimizer.step()

            if not accelerator.optimizer_step_was_skipped:
                lr_scheduler.step()

            global_step += 1

            if accelerator.is_main_process:
                accelerator.print(f"epoch: {epoch}, cureent step: {batch_cnt}, total step: {len(train_dataloader)}, skip:{accelerator.optimizer_step_was_skipped}, loss:{round(train_loss, 3)}, acc:{round(acc, 3)}, length:{len(input_ids[0])}, lr:{lr_scheduler.get_last_lr()[0]}")

            if global_step % 3 == 0 and accelerator.is_main_process:
                writer.add_scalar('skip', int(accelerator.optimizer_step_was_skipped), global_step=global_step)
                writer.add_scalar('loss', train_loss, global_step=global_step)
                writer.add_scalar('acc', acc, global_step=global_step)
                writer.add_scalar('lr', lr_scheduler.get_last_lr()[0], global_step=global_step)

            if global_step % args.eval_step == 0 or global_step == 1:
                torch.cuda.empty_cache()
                model.eval() 

                val_metric = SFTMetric(torch.cuda.current_device())
                for input_ids, attention_mask, labels in val_dataloader:
                    with torch.no_grad():
                        output = model(input_ids=input_ids, attention_mask=attention_mask, labels=labels, return_dict=True)

                    val_metric(output.logits, labels, output.loss)

                val_acc, val_loss = val_metric.get_metric()

                if accelerator.is_main_process:
                    writer.add_scalar(f'val_loss', val_loss, global_step=global_step)
                    writer.add_scalar(f'val_acc', val_acc, global_step=global_step)
                    accelerator.print(f"Epoch: {epoch}, Step: {batch_cnt}, Val loss: {val_loss}, Val acc: {val_acc}")

                model.train()           

            if global_step % args.save_step == 0:
                model.save_checkpoint(args.output_dir, global_step)

    if global_step % args.save_step != 0:
        model.save_checkpoint(args.output_dir, global_step)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Args of sft')

    # Model Args
    parser.add_argument('--model_path', default='./ckpts/moss-16B-base', type=str)
    
    # Data Args
    parser.add_argument('--data_dir', default='./data/sft', type=str)
    parser.add_argument('--output_dir', default='./ckpts/moss-16B-sft', type=str)
    parser.add_argument('--log_dir', default='./train_logs/moss-16B-sft', type=str)
    
    # Training Args
    parser.add_argument('--max_seq_len', default=2048, type=int)
    parser.add_argument('--train_bsz_per_gpu', default=4, type=int)
    parser.add_argument('--eval_bsz_per_gpu', default=4, type=int)
    parser.add_argument('--weight_decay', default=0.1, type=float)
    parser.add_argument('--learning_rate', default=9e-6, type=float)
    parser.add_argument('--warmup_rates', default=0.05, type=int)
    parser.add_argument('--n_epochs', default=2, type=int)

    # Other Args
    parser.add_argument('--save_step', default=3000, type=int)
    parser.add_argument('--eval_step', default=5, type=int)
    parser.add_argument('--seed', default=42, type=int)

    args = parser.parse_args()


    os.makedirs(args.log_dir, exist_ok=True)
    os.makedirs(args.output_dir, exist_ok=True)

    set_seed(args.seed)
    train(args)           



from http import HTTPStatus
from typing import List, Dict
import dashscope
from dashscope.api_entities.dashscope_response import Message
import time
import ray
from byzerllm.utils.types import BlockVLLMStreamServer,StreamOutputs,SingleOutput,SingleOutputMeta
import threading
import asyncio
from byzerllm.utils.langutil import asyncfy_with_semaphore

class CustomSaasAPI:
    def __init__(self, infer_params: Dict[str, str]) -> None:
        self.api_key: str = infer_params["saas.api_key"]  
        self.model = infer_params.get("saas.model", "qwen-turbo")
        self.meta = {
            "model_deploy_type": "saas",
            "backend":"saas",
            "support_stream": True
        } 
        
        self.meta["embedding_mode"] = "embedding"  in  self.model.lower()

        try:
            ray.get_actor("BLOCK_VLLM_STREAM_SERVER")
        except ValueError: 
            try:           
                ray.remote(BlockVLLMStreamServer).options(name="BLOCK_VLLM_STREAM_SERVER",lifetime="detached",max_concurrency=1000).remote()     
            except Exception:
                pass

     # saas/proprietary
    def get_meta(self):
        return [self.meta] 
    
    async def async_get_meta(self):
        return await asyncfy_with_semaphore(self.get_meta)()
    
    def embed_query(self, ins: str, **kwargs): 
        # text-embedding-v1或者text-embedding-v2              
        resp = dashscope.TextEmbedding.call(
        api_key=self.api_key,
        model=self.model,
        input=ins)
        if resp.status_code == HTTPStatus.OK:
            return (resp.output["embeddings"][0]["embedding"],{"metadata":{
                "input_tokens_count":resp.usage["total_tokens"],
                "generated_tokens_count":0}})
        else:
            raise Exception(resp.message)

    async def async_embed_query(self, ins: str, **kwargs):
        return await asyncfy_with_semaphore(self.embed_query)(ins, **kwargs)


    async def async_stream_chat(
            self,
            tokenizer,
            ins: str,
            his: List[dict] = [],
            max_length: int = 1024,
            top_p: float = 0.9,
            temperature: float = 0.1,
            **kwargs
    ):
        
        messages = his + [{"role": "user", "content": ins}]        
        
        start_time = time.monotonic()

        other_params = {}
                
        if "top_k" in kwargs:
            other_params["top_k"] = int(kwargs["top_k"])

        if "stop" in kwargs:
            other_params["stop"] = kwargs["stop"]

        if "enable_search" in kwargs:
            other_params["enable_search"] = kwargs["enable_search"]        

        if "enable_search_enhance" in kwargs:
            other_params["enable_search_enhance"] = kwargs["enable_search_enhance"]

        if "stream" in kwargs:        
            other_params["stream"] = kwargs["stream"]

        if "incremental_output" in kwargs:
            other_params["incremental_output"] = kwargs["incremental_output"]    

        stream = kwargs.get("stream",False)    
        
        res_data = await asyncfy_with_semaphore(lambda:dashscope.Generation.call(model = self.model,
                                            messages=[Message(**message) for message in messages],
                                            api_key=self.api_key,
                                            max_tokens=max_length,
                                            temperature=temperature,
                                            top_p=top_p,
                                            result_format='message',**other_params))()
        
        if stream:
            
            server = ray.get_actor("BLOCK_VLLM_STREAM_SERVER")
            request_id = [None]

            def writer(): 
                for response in res_data:                                        
                    if response.status_code == HTTPStatus.OK:
                        v = response.output.choices[0]['message']['content']                        
                        request_id[0] = response.request_id                        
                        ray.get(server.add_item.remote(request_id[0], 
                                                       StreamOutputs(outputs=[SingleOutput(text=v,metadata=SingleOutputMeta(
                                                           input_tokens_count=response["usage"]["input_tokens"],
                                                           generated_tokens_count=response["usage"]["output_tokens"],
                                                       ))])
                                                       ))
                        
                    else:
                        print('Request id: %s, Status code: %s, error code: %s, error message: %s' % (
                            response.request_id, response.status_code,
                            response.code, response.message
                        ),flush=True) 
                ray.get(server.mark_done.remote(request_id[0]))

            threading.Thread(target=writer,daemon=True).start()            
                               
            time_count= 10*100
            while request_id[0] is None and time_count > 0:
                time.sleep(0.01)
                time_count -= 1
            
            if request_id[0] is None:
                raise Exception("Failed to get request id")
            
            def write_running():
                return ray.get(server.add_item.remote(request_id[0], "RUNNING"))
                        
            await asyncio.to_thread(write_running)
            return [("",{"metadata":{"request_id":request_id[0],"stream_server":"BLOCK_VLLM_STREAM_SERVER"}})]  
              
        time_cost = time.monotonic() - start_time
        
        if res_data["status_code"] == HTTPStatus.OK:
             generated_text = res_data["output"]["choices"][0]["message"]["content"]
             generated_tokens_count = res_data["usage"]["output_tokens"]
             input_tokens_count = res_data["usage"]["input_tokens"]

             return [(generated_text,{"metadata":{
                        "request_id":res_data["request_id"],
                        "input_tokens_count":input_tokens_count,
                        "generated_tokens_count":generated_tokens_count,
                        "time_cost":time_cost,
                        "first_token_time":0,
                        "speed":float(generated_tokens_count)/time_cost,        
                    }})] 
        else:
            s = 'Request id: %s, Status code: %s, error code: %s, error message: %s' % (
                res_data.request_id, res_data.status_code,
                res_data.code, res_data.message
            )
            print(s)
            raise Exception(s)

    
        




import ray
import os
import time
import types
import copy
import asyncio
from typing import Any,Any,Dict, List,Tuple,Generator,Optional,Union
from pyjava.api.mlsql import DataServer
from byzerllm.utils.metrics import Metric
from byzerllm import BlockRow
from byzerllm.utils import (VLLMStreamServer,
                            StreamOutputs,
                            SingleOutput,
                            SingleOutputMeta,
                            compute_max_new_tokens,
                            tokenize_stopping_sequences,
                            ) 
from byzerllm.utils.tokenizer import get_real_tokenizer,get_local_tokenizer,validate_args_engine_use_ray                        
from byzerllm.utils.ray_utils import get_actor_info

try:
    from byzerllm.auto.backend_llama_cpp import LlamaCppBackend
except ImportError:
    print("python_llama_cpp is not installed, if you want to use llama_cpp backend,please install it by `pip install python_llama_cpp`",flush=True)
    pass

try:
    from vllm.engine.async_llm_engine import AsyncLLMEngine,AsyncEngineArgs,_AsyncLLMEngine    
    from vllm import  SamplingParams
    from vllm.utils import random_uuid    
except ImportError:
    print("vllm is not installed, if you want to use vllm backend,please install it by `pip install vllm`",flush=True)
    pass
import torch
from transformers import AutoTokenizer, AutoModelForCausalLM,BitsAndBytesConfig,StoppingCriteriaList,GenerationConfig
from byzerllm.utils.types import StopSequencesCriteria
from ray.util.client.common import ClientActorHandle

INFERENCE_NAME = "auto"
INFER_TOKEN_METRICS = Metric()

def get_bool(params:Dict[str,str],key:str,default:bool=False)->bool:
    if key in params:
        if isinstance(params[key],bool):
            return params[key]
        else:            
            return params[key] == "true" or params[key] == "True"
    return default

def get_int(params:Dict[str,str],key:str,default:int=0)->int:
    if key in params:
        return int(params[key])
    return default

def get_float(params:Dict[str,str],key:str,default:float=0.0)->float:
    if key in params:
        return float(params[key])
    return default

def get_str(params:Dict[str,str],key:str,default:str="")->str:
    if key in params:
        return params[key]
    return default    


def stream_chat(self,tokenizer,ins:str, his:List[Dict[str,str]]=[],  
        max_length:int=4090, 
        top_p:float=0.95,
        temperature:float=0.1,**kwargs):
 
    if self.get_meta()[0]["message_format"]:
        config = copy.deepcopy(self.generation_config)
        config.max_length = max_length
        config.temperature = temperature
        config.top_p = top_p        
        
        if "max_new_tokens" in kwargs:
            config.max_new_tokens = int(kwargs["max_new_tokens"])
        
        conversations = his + [{"content":ins,"role":"user"}]
        start_time = time.monotonic()
        response = self.chat(tokenizer, messages=conversations,generation_config=config)
        time_taken = time.monotonic() - start_time

        generated_tokens_count = tokenizer(response, return_token_type_ids=False,return_tensors="pt")["input_ids"].shape[1]
        print(f"chat took {time_taken} s to complete. tokens/s:{float(generated_tokens_count)/time_taken}",flush=True)
        
        return [(response,{"metadata":{
            "request_id":"",
            "input_tokens_count": -1,
            "generated_tokens_count":generated_tokens_count,
            "time_cost":time_taken,
            "first_token_time": -1.0,
            "speed":float(generated_tokens_count)/time_taken*1000,
            "prob": -1.0
        }})] 

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu") 
    timeout_s = float(kwargs.get("timeout_s",60*5)) 
    skip_check_min_length = int(kwargs.get("stopping_sequences_skip_check_min_length",0))       
    
    tokens = tokenizer(ins, return_token_type_ids=False,return_tensors="pt").to(device)

    stopping_criteria = None


    if "stopping_sequences" in kwargs:        
        stopping_sequences = [torch.tensor(word).to(device) for word in tokenize_stopping_sequences(tokenizer,kwargs["stopping_sequences"].split(","))]    
        input_length = tokens["input_ids"].shape[1]
        stopping_criteria=StoppingCriteriaList([StopSequencesCriteria(
            tokenizer=tokenizer,
            stops=stopping_sequences,
            input_start=input_length,
            skip_check_min_length=skip_check_min_length
            )])

    config = self.config

    max_new_tokens = compute_max_new_tokens(tokens, min(max_length, getattr(config, "model_max_length", max_length))) 

    other_params = {}  
    if "early_stopping" in kwargs:
        other_params["early_stopping"] = bool(kwargs["early_stopping"])

    if "repetition_penalty" in kwargs:
        other_params["repetition_penalty"] = float(kwargs["repetition_penalty"])

    if self.generation_config and self.generation_config.eos_token_id:
        other_params["eos_token_id"] = self.generation_config.eos_token_id

    if self.generation_config and self.generation_config.pad_token_id:
        other_params["pad_token_id"] = self.generation_config.pad_token_id
    
    if self.generation_config and self.generation_config.bos_token_id:
        other_params["bos_token_id"] = self.generation_config.bos_token_id
    
    start_time = time.monotonic()        
    response = self.generate(
        input_ids=tokens["input_ids"],
        max_new_tokens= max_new_tokens,        
        temperature=temperature,
        top_p=top_p,        
        max_time=timeout_s,
        stopping_criteria=stopping_criteria,
        **other_params
    )    
    time_taken = time.monotonic() - start_time    
    new_tokens = response[0][tokens["input_ids"].shape[1]:]
    print(f"generate took {time_taken} s to complete. tokens/s:{len(new_tokens)/time_taken}",flush=True)
    answer = tokenizer.decode(new_tokens, skip_special_tokens=True)
    
    return [(answer,{"metadata":{
            "request_id":"",
            "input_tokens_count": tokens["input_ids"].shape[1],
            "generated_tokens_count":len(new_tokens),
            "time_cost":time_taken,
            "first_token_time": -1.0,
            "speed":float(generated_tokens_count)/time_taken*1000,
            "prob": -1.0
        }})] 


def init_model(model_dir,infer_params:Dict[str,str]={},sys_conf:Dict[str,str]={}): 
    infer_mode = sys_conf.get("infer_backend","transformers")
    quatization = infer_params.get("quatization","false") == "true"  

    if infer_mode == "llama_cpp":       
        model = LlamaCppBackend(model_path=model_dir,infer_params=infer_params,sys_conf=sys_conf)
        return (model,None)

    if infer_mode == "ray/vllm":        
        from byzerllm.auto.backend_vllm import from_vllm_backend
        llm,tokenizer = from_vllm_backend(model_path=model_dir,infer_params=infer_params,sys_conf=sys_conf)
        return (llm,tokenizer)  

    if  infer_mode == "ray/deepspeed":
        from .backend_ds import DeepSpeedInference,ParallelConfig        
        num_gpus = int(sys_conf.get("num_gpus",1))
        model = DeepSpeedInference(ParallelConfig(num_workers=num_gpus,model_dir=model_dir))    
        return (model,None)                     

    pretrained_model_dir = os.path.join(model_dir,"pretrained_model")
    adaptor_model_dir = model_dir
    is_adaptor_model = os.path.exists(pretrained_model_dir)
    
    if not is_adaptor_model:        
        pretrained_model_dir = model_dir

    tokenizer = AutoTokenizer.from_pretrained(pretrained_model_dir,trust_remote_code=True)    

    quatization = infer_params.get("backend.quantization",infer_params.get("quatization", "false"))

    if quatization in ["4", "8", "true",True,4,8]:
        print(f"enable [{quatization}] quatization.", flush=True)
        load_in_8bit = quatization == "8" or quatization == 8
        # default using int4
        quantization_config = BitsAndBytesConfig(
            load_in_4bit=True,
            bnb_4bit_quant_type="nf4",
            bnb_4bit_use_double_quant=False,
            bnb_4bit_compute_dtype=torch.bfloat16,
        )
        if load_in_8bit:
            llm_int8_threshold = infer_params.get("llm_int8_threshold", 6.0)
            quantization_config = BitsAndBytesConfig(
                load_in_8bit=True,
                llm_int8_threshold=llm_int8_threshold,
                llm_int8_skip_modules=None,
                llm_int8_enable_fp32_cpu_offload=False,
                llm_int8_has_fp16_weight=False,
            )
        model = AutoModelForCausalLM.from_pretrained(
            pretrained_model_dir,
            trust_remote_code=True,
            device_map="auto",
            quantization_config=quantization_config,
        )

    else:
        model = AutoModelForCausalLM.from_pretrained(pretrained_model_dir,trust_remote_code=True,
                                                device_map='auto',                                                
                                                torch_dtype=torch.bfloat16                                                
                                                )
    if is_adaptor_model:
        from peft import PeftModel
        model = PeftModel.from_pretrained(model, adaptor_model_dir)
    
    model.generation_config = GenerationConfig.from_pretrained(pretrained_model_dir)
    
    if "use_cache" in infer_params or "backend.use_cache" in infer_params:
        use_cache = infer_params.get("backend.use_cache",infer_params.get("use_cache", "false"))
        if isinstance(use_cache,bool):
            model.generation_config.use_cache = use_cache
        else:
            model.generation_config.use_cache = use_cache == "true"    
    
    model.eval()  
    if quatization:
        model = torch.compile(model)
           
    has_chat = hasattr(model,"chat")
    
    if "message_format" in infer_params or "backend.message_format" in infer_params:
        message_format = infer_params.get("backend.message_format",infer_params.get("message_format", "false"))
        if isinstance(message_format,bool):
            has_chat = message_format
        else:
            has_chat = message_format == "true"

    extra_meta = {}
    if has_chat:
        extra_meta["message_format"] = True

    def get_meta(self): 
        config = self.config           
        return [{
            "model_deploy_type": "proprietary",
            "backend":"transformers",
            "max_model_len":getattr(config, "model_max_length", -1),
            "architectures":getattr(config, "architectures", []),
            **extra_meta
        }]    

    model.stream_chat = types.MethodType(stream_chat, model)
    model.get_meta = types.MethodType(get_meta, model)     
    return (model,tokenizer)


def sft_train(data_refs:List[DataServer],
              train_params:Dict[str,str],
              conf: Dict[str, str])->Generator[BlockRow,Any,Any]:
    from ..utils.sft import sft_train as common_sft_train
    return common_sft_train(data_refs,train_params,conf) 




from ..conversable_agent import ConversableAgent
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from ....utils.client import ByzerLLM
from byzerllm.utils.retrieval import ByzerRetrieval

class PlannerAgent(ConversableAgent):
    SYSTEM_PROMPT = '''You are a helpful AI assistant. You suggest coding and reasoning steps for another AI assistant to accomplish a task. Do not suggest concrete code. For any action beyond writing code or reasoning, convert it to a step that can be implemented by writing code. For example, browsing the web can be implemented by writing code that reads and prints the content of a web page. Finally, inspect the execution result. If the plan is not good, suggest a better plan. If the execution is wrong, analyze the error and suggest a fix.
'''    

    def __init__(
        self,
        name: str,
        llm: ByzerLLM,        
        retrieval: ByzerRetrieval,                
        system_message: Optional[str] = SYSTEM_PROMPT,        
        is_termination_msg: Optional[Callable[[Dict], bool]] = None,
        max_consecutive_auto_reply: Optional[int] = None,
        human_input_mode: Optional[str] = "NEVER",
        code_execution_config: Optional[Union[Dict, bool]] = False,
        **kwargs,
    ):       
        super().__init__(
            name,
            llm,retrieval,
            system_message,
            is_termination_msg,
            max_consecutive_auto_reply,
            human_input_mode,
            code_execution_config=code_execution_config,            
            **kwargs,
        )
        pass



try:
    from unittest.mock import patch
    def mock_nltk_data_find(resource_name, paths=None):
        return ""
    def mock_nltk_download(resource_name, download_dir=None):
        pass
    patcher_find = patch("nltk.data.find", side_effect=mock_nltk_data_find)
    patcher_download = patch("nltk.download", side_effect=mock_nltk_download)
    patcher_find.start()
    patcher_download.start()
except Exception as e:
    print("Error in patching nltk.download", e)
    pass


import os
from typing import Any, Callable, Dict, Optional, Sequence

from llama_index.core.bridge.pydantic import Field, PrivateAttr

from llama_index.core.llms import (    
    ChatMessage,
    ChatResponse,
    ChatResponseGen,
    CompletionResponse,
    CompletionResponseGen,
    LLMMetadata,
)
from llama_index.core.llms.callbacks import llm_chat_callback, llm_completion_callback
from llama_index.core.llms.custom import CustomLLM
from llama_index.core.llms.custom import (
    completion_response_to_chat_response,
    stream_completion_response_to_chat_response,
)
from llama_index.core.types import BaseOutputParser, PydanticProgramMode
from llama_index.core.utils import get_cache_dir
from byzerllm.utils.client import ByzerLLM

class ByzerAI(CustomLLM):
    """
    ByzerAI is a custom LLM that uses the ByzerLLM API to generate text.
    """    
   
    verbose: bool = Field(
        default=False,
        description="Whether to print verbose output.",
    )

    _model: ByzerLLM = PrivateAttr()

    def __init__(
        self,
        llm:ByzerLLM
    ) -> None:        
        self._model = llm                
        super().__init__()

    @classmethod
    def class_name(cls) -> str:
        return "ByzerAI_llm"

    @property
    def metadata(self) -> LLMMetadata:
        """LLM metadata."""
        return LLMMetadata.parse_obj(self._model.metadata.model_dump())

    @llm_chat_callback()
    def chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        conversations = [{
            "role":message.role,
            "content":message.content
        } for message in messages]
        m = self._model.chat_oai(conversations=conversations)
        completion_response = CompletionResponse(text=m[0].output, raw=None)
        return completion_response_to_chat_response(completion_response)

    @llm_chat_callback()
    def stream_chat(
        self, messages: Sequence[ChatMessage], **kwargs: Any
    ) -> ChatResponseGen:        
        conversations = [{
            "role":message.role,
            "content":message.content
        } for message in messages]
        m = self._model.stream_chat_oai(conversations=conversations)
        def gen():
            v = ""
            for response in m:                
                text:str = response[0]
                metadata:Dict[str,Any] = response[1]
                completion_response = CompletionResponse(text=text, delta=text[len(v):], raw=None)
                v = text
                yield completion_response
        return stream_completion_response_to_chat_response(gen())

    @llm_completion_callback()
    def complete(
        self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponse:        
        m = self._model.chat_oai(conversations=[{"role":"user","content":prompt}])
        completion_response = CompletionResponse(text=m[0].output, raw=None)
        return completion_response

    @llm_completion_callback()
    def stream_complete(
        self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponseGen:
        conversations=[{"role":"user","content":prompt}]
        m = self._model.stream_chat_oai(conversations=conversations)
        def gen():
            v = ""
            for response in m:                
                text:str = response[0]
                metadata:Dict[str,Any] = response[1]
                completion_response = CompletionResponse(text=text, delta=text[len(v):], raw=None)
                v = text
                yield completion_response
        return gen()        